"""Test method in BaseTable class."""

import pytest

from terminaltables3.base_table import BaseTable
from terminaltables3.width_and_alignment import max_dimensions

SINGLE_LINE = (
    ("Name", "Color", "Type"),
    ("Avocado", "green", "nut"),
    ("Tomato", "red", "fruit"),
    ("Lettuce", "green", "vegetable"),
)


@pytest.mark.parametrize("inner_column_border", [True, False])
@pytest.mark.parametrize("style", ["top", "bottom"])
def test_top_bottom(inner_column_border, style):
    """Test top and bottom borders.

    :param bool inner_column_border: Passed to table class.
    :param str style: Passed to method.
    """
    table = BaseTable(SINGLE_LINE, "Example")
    table.inner_column_border = inner_column_border
    outer_widths = max_dimensions(
        table.table_data, table.padding_left, table.padding_right
    )[2]

    # Determine expected.
    if style == "top" and inner_column_border:
        expected = "+Example--+-------+-----------+"
    elif style == "top":
        expected = "+Example--------------------+"
    elif style == "bottom" and inner_column_border:
        expected = "+---------+-------+-----------+"
    else:
        expected = "+---------------------------+"

    # Test.
    actual = "".join(table.horizontal_border(style, outer_widths))
    assert actual == expected


@pytest.mark.parametrize("inner_column_border", [True, False])
@pytest.mark.parametrize("outer_border", [True, False])
@pytest.mark.parametrize("style", ["heading", "footing"])
def test_heading_footing(inner_column_border, outer_border, style):
    """Test heading and footing borders.

    :param bool inner_column_border: Passed to table class.
    :param bool outer_border: Passed to table class.
    :param str style: Passed to method.
    """
    table = BaseTable(SINGLE_LINE)
    table.inner_column_border = inner_column_border
    table.outer_border = outer_border
    outer_widths = max_dimensions(
        table.table_data, table.padding_left, table.padding_right
    )[2]

    # Determine expected.
    if style == "heading" and outer_border:
        expected = (
            "+---------+-------+-----------+"
            if inner_column_border
            else "+---------------------------+"
        )
    elif style == "heading":
        expected = (
            "---------+-------+-----------"
            if inner_column_border
            else "---------------------------"
        )
    elif style == "footing" and outer_border:
        expected = (
            "+---------+-------+-----------+"
            if inner_column_border
            else "+---------------------------+"
        )
    else:
        expected = (
            "---------+-------+-----------"
            if inner_column_border
            else "---------------------------"
        )

    # Test.
    actual = "".join(table.horizontal_border(style, outer_widths))
    assert actual == expected


@pytest.mark.parametrize("inner_column_border", [True, False])
@pytest.mark.parametrize("outer_border", [True, False])
def test_row(inner_column_border, outer_border):
    """Test inner borders.

    :param bool inner_column_border: Passed to table class.
    :param bool outer_border: Passed to table class.
    """
    table = BaseTable(SINGLE_LINE)
    table.inner_column_border = inner_column_border
    table.outer_border = outer_border
    outer_widths = max_dimensions(
        table.table_data, table.padding_left, table.padding_right
    )[2]

    # Determine expected.
    if inner_column_border and outer_border:
        expected = "+---------+-------+-----------+"
    elif inner_column_border:
        expected = "---------+-------+-----------"
    elif outer_border:
        expected = "+---------------------------+"
    else:
        expected = "---------------------------"

    # Test.
    actual = "".join(table.horizontal_border("row", outer_widths))
    assert actual == expected



"""Test AsciiTable class."""

import pytest

from terminaltables3.other_tables import AsciiTable

SINGLE_LINE = (
    ("Name", "Color", "Type"),
    ("Avocado", "green", "nut"),
    ("Tomato", "red", "fruit"),
    ("Lettuce", "green", "vegetable"),
)

MULTI_LINE = (
    ("Show", "Characters"),
    (
        "Rugrats",
        "Tommy Pickles, Chuckie Finster, Phillip DeVille, Lillian DeVille, Angelica Pickles,\nDil Pickles",
    ),
    ("South Park", "Stan Marsh, Kyle Broflovski, Eric Cartman, Kenny McCormick"),
)


@pytest.fixture(autouse=True)
def patch(monkeypatch):
    """Monkeypatch before every test function in this module.

    :param monkeypatch: pytest fixture.
    """
    monkeypatch.setattr("terminaltables3.ascii_table.terminal_size", lambda: (79, 24))
    monkeypatch.setattr(
        "terminaltables3.width_and_alignment.terminal_size", lambda: (79, 24)
    )


@pytest.mark.parametrize(
    "table_data,column_number,expected",
    [
        ([], 0, IndexError),
        ([[]], 0, IndexError),
        ([[""]], 1, IndexError),
        (SINGLE_LINE, 0, 55),
        (SINGLE_LINE, 1, 53),
        (SINGLE_LINE, 2, 57),
        (MULTI_LINE, 0, -11),
        (MULTI_LINE, 1, 62),
    ],
)
def test_column_max_width(table_data, column_number, expected):
    """Test method in class.

    :param iter table_data: Passed to AsciiTable.__init__().
    :param int column_number: Passed to AsciiTable.column_max_width().
    :param int expected: Expected return value of AsciiTable.column_max_width().
    """
    table = AsciiTable(table_data)

    if expected is IndexError:
        with pytest.raises(IndexError):
            table.column_max_width(column_number)
        return

    actual = table.column_max_width(column_number)
    assert actual == expected


def test_column_widths():
    """Test method in class."""
    assert AsciiTable([]).column_widths == []

    table = AsciiTable(SINGLE_LINE)
    actual = table.column_widths
    assert actual == [7, 5, 9]


@pytest.mark.parametrize(
    "table_data,terminal_width,expected",
    [
        ([], None, True),
        ([[]], None, True),
        ([[""]], None, True),
        (SINGLE_LINE, None, True),
        (SINGLE_LINE, 30, False),
        (MULTI_LINE, None, False),
        (MULTI_LINE, 100, True),
    ],
)
def test_ok(monkeypatch, table_data, terminal_width, expected):
    """Test method in class.

    :param monkeypatch: pytest fixture.
    :param iter table_data: Passed to AsciiTable.__init__().
    :param int terminal_width: Monkeypatch width of terminal_size() if not None.
    :param bool expected: Expected return value.
    """
    if terminal_width is not None:
        monkeypatch.setattr(
            "terminaltables3.ascii_table.terminal_size", lambda: (terminal_width, 24)
        )
    table = AsciiTable(table_data)
    actual = table.ok
    assert actual is expected


@pytest.mark.parametrize(
    "table_data,expected",
    [
        ([], 2),
        ([[]], 2),
        ([[""]], 4),
        ([[" "]], 5),
        (SINGLE_LINE, 31),
        (MULTI_LINE, 100),
    ],
)
def test_table_width(table_data, expected):
    """Test method in class.

    :param iter table_data: Passed to AsciiTable.__init__().
    :param int expected: Expected return value.
    """
    table = AsciiTable(table_data)
    actual = table.table_width
    assert actual == expected



"""Sphinx configuration file."""

import os
import sys
import time


# General configuration.
sys.path.append(os.path.realpath(os.path.join(os.path.dirname(__file__), "..")))
author = "@Robpol86"
copyright = "{}, {}".format(time.strftime("%Y"), author)
html_last_updated_fmt = f"%c {time.tzname[time.localtime().tm_isdst]}"
master_doc = "index"
project = __import__("setup").NAME
pygments_style = "friendly"
release = version = __import__("setup").VERSION
templates_path = ["_templates"]
extensions = []


# Options for HTML output.
html_context = dict(
    conf_py_path="/docs/",
    display_github=True,
    github_repo=os.environ.get("TRAVIS_REPO_SLUG", "/" + project).split("/", 1)[1],
    github_user=os.environ.get("TRAVIS_REPO_SLUG", "robpol86/").split("/", 1)[0],
    github_version=os.environ.get("TRAVIS_BRANCH", "master"),
    source_suffix=".rst",
)
html_copy_source = False
html_favicon = "favicon.ico"
html_theme = "sphinx_rtd_theme"
html_title = project


# autodoc
extensions.append("sphinx.ext.autodoc")


# extlinks.
extensions.append("sphinx.ext.extlinks")
extlinks = {
    "github": (
        f"https://github.com/robpol86/{project}/blob/v{version}/%s",
        "",
    )
}


# google analytics
extensions.append("sphinxcontrib.googleanalytics")
googleanalytics_id = "UA-82627369-1"


# SCVersioning.
scv_banner_greatest_tag = True
scv_grm_exclude = (".gitignore", ".nojekyll", "README.rst")
scv_overflow = ("-W",)
scv_show_banner = True
scv_sort = ("semver", "time")



"""Take screenshots and search for subimages in images."""

import ctypes
import os
import random
import struct
import subprocess
import time

try:
    from itertools import izip
except ImportError:
    izip = zip  # Py3

from tests import PROJECT_ROOT

STARTF_USESHOWWINDOW = getattr(subprocess, "STARTF_USESHOWWINDOW", 1)
STILL_ACTIVE = 259
SW_MAXIMIZE = 3


class StartupInfo(ctypes.Structure):
    """STARTUPINFO structure."""

    _fields_ = [
        ("cb", ctypes.c_ulong),
        ("lpReserved", ctypes.c_char_p),
        ("lpDesktop", ctypes.c_char_p),
        ("lpTitle", ctypes.c_char_p),
        ("dwX", ctypes.c_ulong),
        ("dwY", ctypes.c_ulong),
        ("dwXSize", ctypes.c_ulong),
        ("dwYSize", ctypes.c_ulong),
        ("dwXCountChars", ctypes.c_ulong),
        ("dwYCountChars", ctypes.c_ulong),
        ("dwFillAttribute", ctypes.c_ulong),
        ("dwFlags", ctypes.c_ulong),
        ("wShowWindow", ctypes.c_ushort),
        ("cbReserved2", ctypes.c_ushort),
        ("lpReserved2", ctypes.c_char_p),
        ("hStdInput", ctypes.c_ulong),
        ("hStdOutput", ctypes.c_ulong),
        ("hStdError", ctypes.c_ulong),
    ]

    def __init__(self, maximize=False, title=None):
        """Constructor.

        :param bool maximize: Start process in new console window, maximized.
        :param bytes title: Set new window title to this instead of exe path.
        """
        super().__init__()
        self.cb = ctypes.sizeof(self)
        if maximize:
            self.dwFlags |= STARTF_USESHOWWINDOW
            self.wShowWindow = SW_MAXIMIZE
        if title:
            self.lpTitle = ctypes.c_char_p(title)


class ProcessInfo(ctypes.Structure):
    """PROCESS_INFORMATION structure."""

    _fields_ = [
        ("hProcess", ctypes.c_void_p),
        ("hThread", ctypes.c_void_p),
        ("dwProcessId", ctypes.c_ulong),
        ("dwThreadId", ctypes.c_ulong),
    ]


class RunNewConsole:
    """Run the command in a new console window. Windows only. Use in a with statement.

    subprocess sucks and really limits your access to the win32 API. Its implementation is half-assed. Using this so
    that STARTUPINFO.lpTitle actually works and STARTUPINFO.dwFillAttribute produce the expected result.
    """

    def __init__(self, command, maximized=False, title=None):
        """Constructor.

        :param iter command: Command to run.
        :param bool maximized: Start process in new console window, maximized.
        :param bytes title: Set new window title to this. Needed by user32.FindWindow.
        """
        if title is None:
            title = "pytest-{}-{}".format(
                os.getpid(), random.randint(1000, 9999)
            ).encode("ascii")
        self.startup_info = StartupInfo(maximize=maximized, title=title)
        self.process_info = ProcessInfo()
        self.command_str = subprocess.list2cmdline(command).encode("ascii")
        self._handles = []
        self._kernel32 = ctypes.LibraryLoader(ctypes.WinDLL).kernel32
        self._kernel32.GetExitCodeProcess.argtypes = [
            ctypes.c_void_p,
            ctypes.POINTER(ctypes.c_ulong),
        ]
        self._kernel32.GetExitCodeProcess.restype = ctypes.c_long

    def __del__(self):
        """Close win32 handles."""
        while self._handles:
            try:
                self._kernel32.CloseHandle(
                    self._handles.pop(0)
                )  # .pop() is thread safe.
            except IndexError:
                break

    def __enter__(self):
        """Entering the `with` block. Runs the process."""
        if not self._kernel32.CreateProcessA(
            None,  # lpApplicationName
            self.command_str,  # lpCommandLine
            None,  # lpProcessAttributes
            None,  # lpThreadAttributes
            False,  # bInheritHandles
            subprocess.CREATE_NEW_CONSOLE,  # dwCreationFlags
            None,  # lpEnvironment
            str(PROJECT_ROOT).encode("ascii"),  # lpCurrentDirectory
            ctypes.byref(self.startup_info),  # lpStartupInfo
            ctypes.byref(self.process_info),  # lpProcessInformation
        ):
            raise ctypes.WinError()

        # Add handles added by the OS.
        self._handles.append(self.process_info.hProcess)
        self._handles.append(self.process_info.hThread)

        # Get hWnd.
        self.hwnd = 0
        for _ in range(int(5 / 0.1)):
            # Takes time for console window to initialize.
            self.hwnd = ctypes.windll.user32.FindWindowA(
                None, self.startup_info.lpTitle
            )
            if self.hwnd:
                break
            time.sleep(0.1)
        assert self.hwnd

        # Return generator that yields window size/position.
        return self._iter_pos()

    def __exit__(self, *_):
        """Cleanup."""
        try:
            # Verify process exited 0.
            status = ctypes.c_ulong(STILL_ACTIVE)
            while status.value == STILL_ACTIVE:
                time.sleep(0.1)
                if not self._kernel32.GetExitCodeProcess(
                    self.process_info.hProcess, ctypes.byref(status)
                ):
                    raise ctypes.WinError()
            assert status.value == 0
        finally:
            # Close handles.
            self.__del__()

    def _iter_pos(self):
        """Yield new console window's current position and dimensions.

        :return: Yields region the new window is in (left, upper, right, lower).
        :rtype: tuple
        """
        rect = ctypes.create_string_buffer(
            16
        )  # To be written to by GetWindowRect. RECT structure.
        while ctypes.windll.user32.GetWindowRect(self.hwnd, rect):
            left, top, right, bottom = struct.unpack("llll", rect.raw)
            width, height = right - left, bottom - top
            assert width > 1
            assert height > 1
            yield left, top, right, bottom
        raise StopIteration


def iter_rows(pil_image):
    """Yield tuple of pixels for each row in the image.

    itertools.izip in Python 2.x and zip in Python 3.x are writen in C. Much faster than anything else I've found
    written in pure Python.

    From:
    http://stackoverflow.com/questions/1624883/alternative-way-to-split-a-list-into-groups-of-n/1625023#1625023

    :param PIL.Image.Image pil_image: Image to read from.

    :return: Yields rows.
    :rtype: tuple
    """
    iterator = izip(*(iter(pil_image.getdata()),) * pil_image.width)
    yield from iterator


def get_most_interesting_row(pil_image):
    """Look for a row in the image that has the most unique pixels.

    :param PIL.Image.Image pil_image: Image to read from.

    :return: Row (tuple of pixel tuples), row as a set, first pixel tuple, y offset from top.
    :rtype: tuple
    """
    final = (None, set(), None, None)  # row, row_set, first_pixel, y_pos
    for y_pos, row in enumerate(iter_rows(pil_image)):
        row_set = set(row)
        if len(row_set) > len(final[1]):
            final = row, row_set, row[0], y_pos
        if len(row_set) == pil_image.width:
            break  # Can't get bigger.
    return final


def count_subimages(screenshot, subimg):
    """Check how often subimg appears in the screenshot image.

    :param PIL.Image.Image screenshot: Screen shot to search through.
    :param PIL.Image.Image subimg: Subimage to search for.

    :return: Number of times subimg appears in the screenshot.
    :rtype: int
    """
    # Get row to search for.
    si_pixels = list(subimg.getdata())  # Load entire subimg into memory.
    si_width = subimg.width
    si_height = subimg.height
    si_row, si_row_set, si_pixel, si_y = get_most_interesting_row(subimg)
    occurrences = 0

    # Look for subimg row in screenshot, then crop and compare pixel arrays.
    for y_pos, row in enumerate(iter_rows(screenshot)):
        if si_row_set - set(row):
            continue  # Some pixels not found.
        for x_pos in range(screenshot.width - si_width + 1):
            if row[x_pos] != si_pixel:
                continue  # First pixel does not match.
            if row[x_pos : x_pos + si_width] != si_row:
                continue  # Row does not match.
            # Found match for interesting row of subimg in screenshot.
            y_corrected = y_pos - si_y
            with screenshot.crop(
                (x_pos, y_corrected, x_pos + si_width, y_corrected + si_height)
            ) as cropped:
                if list(cropped.getdata()) == si_pixels:
                    occurrences += 1

    return occurrences


def try_candidates(screenshot, subimg_candidates, expected_count):
    """Call count_subimages() for each subimage candidate until.

    If you get ImportError run "pip install pillow". Only OSX and Windows is supported.

    :param PIL.Image.Image screenshot: Screen shot to search through.
    :param iter subimg_candidates: Subimage paths to look for. List of strings.
    :param int expected_count: Try until any a subimage candidate is found this many times.

    :return: Number of times subimg appears in the screenshot.
    :rtype: int
    """
    from PIL import Image

    count_found = 0

    for subimg_path in subimg_candidates:
        with Image.open(subimg_path) as rgba_s:
            with rgba_s.convert(mode="RGB") as subimg:
                # Make sure subimage isn't too large.
                assert subimg.width < 256
                assert subimg.height < 256

                # Count.
                count_found = count_subimages(screenshot, subimg)
                if count_found == expected_count:
                    break  # No need to try other candidates.

    return count_found


def screenshot_until_match(save_to, timeout, subimg_candidates, expected_count, gen):
    """Take screenshots until one of the 'done' subimages is found. Image is saved when subimage found or at timeout.

    If you get ImportError run "pip install pillow". Only OSX and Windows is supported.

    :param str save_to: Save screenshot to this PNG file path when expected count found or timeout.
    :param int timeout: Give up after these many seconds.
    :param iter subimg_candidates: Subimage paths to look for. List of strings.
    :param int expected_count: Keep trying until any of subimg_candidates is found this many times.
    :param iter gen: Generator yielding window position and size to crop screenshot to.
    """
    from PIL import ImageGrab

    assert save_to.endswith(".png")
    stop_after = time.time() + timeout

    # Take screenshots until subimage is found.
    while True:
        with ImageGrab.grab(next(gen)) as rgba:
            with rgba.convert(mode="RGB") as screenshot:
                count_found = try_candidates(
                    screenshot, subimg_candidates, expected_count
                )
                if count_found == expected_count or time.time() > stop_after:
                    screenshot.save(save_to)
                    assert count_found == expected_count
                    return
        time.sleep(0.5)



#!/usr/bin/env python
"""Simple example usage of terminaltables and column_max_width().

Just prints sample text and exits.
"""


from textwrap import wrap

from terminaltables3 import SingleTable

LONG_STRING = (
    "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do eiusmod tempor incididunt ut labore "
    "et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut "
    "aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum "
    "dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui "
    "officia deserunt mollit anim id est laborum."
)


def main():
    """Main function."""
    table_data = [
        [
            "Long String",
            "",
        ],  # One row. Two columns. Long string will replace this empty string.
    ]
    table = SingleTable(table_data)

    # Calculate newlines.
    max_width = table.column_max_width(1)
    wrapped_string = "\n".join(wrap(LONG_STRING, max_width))
    table.table_data[0][1] = wrapped_string

    print(table.table)


if __name__ == "__main__":
    main()



"""GithubFlavoredMarkdownTable end to end testing."""

from terminaltables3 import GithubFlavoredMarkdownTable


def test_single_line():
    """Test single-lined cells."""
    table_data = [
        ["Name", "Color", "Type"],
        ["Avocado", "green", "nut"],
        ["Tomato", "red", "fruit"],
        ["Lettuce", "green", "vegetable"],
        ["Watermelon", "green"],
        [],
    ]
    table = GithubFlavoredMarkdownTable(table_data)
    table.inner_footing_row_border = True
    table.justify_columns[0] = "left"
    table.justify_columns[1] = "center"
    table.justify_columns[2] = "right"
    actual = table.table

    expected = (
        "| Name       | Color |      Type |\n"
        "|:-----------|:-----:|----------:|\n"
        "| Avocado    | green |       nut |\n"
        "| Tomato     |  red  |     fruit |\n"
        "| Lettuce    | green | vegetable |\n"
        "| Watermelon | green |           |\n"
        "|            |       |           |"
    )
    assert actual == expected


def test_multi_line():
    """Test multi-lined cells."""
    table_data = [
        ["Show", "Characters"],
        [
            "Rugrats",
            "Tommy Pickles, Chuckie Finster, Phillip DeVille, Lillian DeVille, Angelica Pickles,\nDil Pickles",
        ],
        ["South Park", "Stan Marsh, Kyle Broflovski, Eric Cartman, Kenny McCormick"],
    ]
    table = GithubFlavoredMarkdownTable(table_data)

    # Test defaults.
    actual = table.table
    expected = (
        "| Show       | Characters                                                                          |\n"
        "|------------|-------------------------------------------------------------------------------------|\n"
        "| Rugrats    | Tommy Pickles, Chuckie Finster, Phillip DeVille, Lillian DeVille, Angelica Pickles, |\n"
        "|            | Dil Pickles                                                                         |\n"
        "| South Park | Stan Marsh, Kyle Broflovski, Eric Cartman, Kenny McCormick                          |"
    )
    assert actual == expected

    # Test inner row border.
    table.inner_row_border = True
    actual = table.table
    expected = (
        "| Show       | Characters                                                                          |\n"
        "|------------|-------------------------------------------------------------------------------------|\n"
        "| Rugrats    | Tommy Pickles, Chuckie Finster, Phillip DeVille, Lillian DeVille, Angelica Pickles, |\n"
        "|            | Dil Pickles                                                                         |\n"
        "| South Park | Stan Marsh, Kyle Broflovski, Eric Cartman, Kenny McCormick                          |"
    )
    assert actual == expected

    # Justify right.
    table.justify_columns = {1: "right"}
    actual = table.table
    expected = (
        "| Show       |                                                                          Characters |\n"
        "|------------|------------------------------------------------------------------------------------:|\n"
        "| Rugrats    | Tommy Pickles, Chuckie Finster, Phillip DeVille, Lillian DeVille, Angelica Pickles, |\n"
        "|            |                                                                         Dil Pickles |\n"
        "| South Park |                          Stan Marsh, Kyle Broflovski, Eric Cartman, Kenny McCormick |"
    )
    assert actual == expected



"""Test function in module."""

from terminaltables3.build import flatten


def test_one_line():
    """Test with one line cells."""
    table = [
        [">", "Left Cell", "|", "Center Cell", "|", "Right Cell", "<"],
    ]
    actual = flatten(table)
    expected = ">Left Cell|Center Cell|Right Cell<"
    assert actual == expected


def test_two_line():
    """Test with two line cells."""
    table = [
        [">", "Left ", "|", "Center", "|", "Right", "<"],
        [">", "Cell1", "|", "Cell2 ", "|", "Cell3", "<"],
    ]
    actual = flatten(table)
    expected = ">Left |Center|Right<\n" ">Cell1|Cell2 |Cell3<"
    assert actual == expected



"""Test function in module."""

import pytest
from colorama import Fore
from colorclass import Color
from termcolor import colored

from terminaltables3.width_and_alignment import visible_width


@pytest.mark.parametrize(
    "string,expected",
    [
        # str
        ("hello, world", 12),
        ("世界你好", 8),
        ("蓝色", 4),
        ("שלום", 4),
        ("معرب", 4),
        ("hello 世界", 10),
        # str+ansi
        ("\x1b[34mhello, world\x1b[39m", 12),
        ("\x1b[34m世界你好\x1b[39m", 8),
        ("\x1b[34m蓝色\x1b[39m", 4),
        ("\x1b[34mשלום\x1b[39m", 4),
        ("\x1b[34mمعرب\x1b[39m", 4),
        ("\x1b[34mhello 世界\x1b[39m", 10),
        # colorclass
        (Color("{blue}hello, world{/blue}"), 12),
        (Color("{blue}世界你好{/blue}"), 8),
        (Color("{blue}蓝色{/blue}"), 4),
        (Color("{blue}שלום{/blue}"), 4),
        (Color("{blue}معرب{/blue}"), 4),
        (Color("{blue}hello 世界{/blue}"), 10),
        # colorama
        (Fore.BLUE + "hello, world" + Fore.RESET, 12),
        (Fore.BLUE + "世界你好" + Fore.RESET, 8),
        (Fore.BLUE + "蓝色" + Fore.RESET, 4),
        (Fore.BLUE + "שלום" + Fore.RESET, 4),
        (Fore.BLUE + "معرب" + Fore.RESET, 4),
        (Fore.BLUE + "hello 世界" + Fore.RESET, 10),
        # termcolor
        (colored("hello, world", "blue"), 12),
        (colored("世界你好", "blue"), 8),
        (colored("蓝色", "blue"), 4),
        (colored("שלום", "blue"), 4),
        (colored("معرب", "blue"), 4),
        (colored("hello 世界", "blue"), 10),
    ],
)
def test(string, expected):
    """Test function with different color libraries.

    :param str string: Input string to measure.
    :param int expected: Expected visible width of string (some characters are len() == 1 but take up 2 spaces).
    """
    assert visible_width(string) == expected



"""PorcelainTable end to end testing."""

from terminaltables3 import PorcelainTable


def test_single_line():
    """Test single-lined cells."""
    table_data = [
        ["Name", "Color", "Type"],
        ["Avocado", "green", "nut"],
        ["Tomato", "red", "fruit"],
        ["Lettuce", "green", "vegetable"],
        ["Watermelon", "green"],
    ]
    table = PorcelainTable(table_data)
    table.justify_columns[0] = "left"
    table.justify_columns[1] = "center"
    table.justify_columns[2] = "right"
    actual = table.table

    expected = (
        " Name       | Color |      Type \n"
        " Avocado    | green |       nut \n"
        " Tomato     |  red  |     fruit \n"
        " Lettuce    | green | vegetable \n"
        " Watermelon | green |           "
    )
    assert actual == expected


def test_multi_line():
    """Test multi-lined cells."""
    table_data = [
        ["Show", "Characters"],
        [
            "Rugrats",
            "Tommy Pickles, Chuckie Finster, Phillip DeVille, Lillian DeVille, Angelica Pickles,\nDil Pickles",
        ],
        ["South Park", "Stan Marsh, Kyle Broflovski, Eric Cartman, Kenny McCormick"],
    ]
    table = PorcelainTable(table_data)

    # Test defaults.
    actual = table.table
    expected = (
        " Show       | Characters                                                                          \n"
        " Rugrats    | Tommy Pickles, Chuckie Finster, Phillip DeVille, Lillian DeVille, Angelica Pickles, \n"
        "            | Dil Pickles                                                                         \n"
        " South Park | Stan Marsh, Kyle Broflovski, Eric Cartman, Kenny McCormick                          "
    )
    assert actual == expected

    # Justify right.
    table.justify_columns = {1: "right"}
    actual = table.table
    expected = (
        " Show       |                                                                          Characters \n"
        " Rugrats    | Tommy Pickles, Chuckie Finster, Phillip DeVille, Lillian DeVille, Angelica Pickles, \n"
        "            |                                                                         Dil Pickles \n"
        " South Park |                          Stan Marsh, Kyle Broflovski, Eric Cartman, Kenny McCormick "
    )
    assert actual == expected



"""AsciiTable end to end testing."""

import sys
from textwrap import dedent

import py
import pytest

from terminaltables3 import AsciiTable
from terminaltables3.terminal_io import IS_WINDOWS
from tests import PROJECT_ROOT
from tests.screenshot import RunNewConsole, screenshot_until_match

HERE = py.path.local(__file__).dirpath()


def test_single_line():
    """Test single-lined cells."""
    table_data = [
        ["Name", "Color", "Type"],
        ["Avocado", "green", "nut"],
        ["Tomato", "red", "fruit"],
        ["Lettuce", "green", "vegetable"],
        ["Watermelon", "green"],
        [],
    ]
    table = AsciiTable(table_data, "Example")
    table.inner_footing_row_border = True
    table.justify_columns[0] = "left"
    table.justify_columns[1] = "center"
    table.justify_columns[2] = "right"
    actual = table.table

    expected = (
        "+Example-----+-------+-----------+\n"
        "| Name       | Color |      Type |\n"
        "+------------+-------+-----------+\n"
        "| Avocado    | green |       nut |\n"
        "| Tomato     |  red  |     fruit |\n"
        "| Lettuce    | green | vegetable |\n"
        "| Watermelon | green |           |\n"
        "+------------+-------+-----------+\n"
        "|            |       |           |\n"
        "+------------+-------+-----------+"
    )
    assert actual == expected


def test_multi_line():
    """Test multi-lined cells."""
    table_data = [
        ["Show", "Characters"],
        [
            "Rugrats",
            "Tommy Pickles, Chuckie Finster, Phillip DeVille, Lillian DeVille, Angelica Pickles,\nDil Pickles",
        ],
        ["South Park", "Stan Marsh, Kyle Broflovski, Eric Cartman, Kenny McCormick"],
    ]
    table = AsciiTable(table_data)

    # Test defaults.
    actual = table.table
    expected = (
        "+------------+-------------------------------------------------------------------------------------+\n"
        "| Show       | Characters                                                                          |\n"
        "+------------+-------------------------------------------------------------------------------------+\n"
        "| Rugrats    | Tommy Pickles, Chuckie Finster, Phillip DeVille, Lillian DeVille, Angelica Pickles, |\n"
        "|            | Dil Pickles                                                                         |\n"
        "| South Park | Stan Marsh, Kyle Broflovski, Eric Cartman, Kenny McCormick                          |\n"
        "+------------+-------------------------------------------------------------------------------------+"
    )
    assert actual == expected

    # Test inner row border.
    table.inner_row_border = True
    actual = table.table
    expected = (
        "+------------+-------------------------------------------------------------------------------------+\n"
        "| Show       | Characters                                                                          |\n"
        "+------------+-------------------------------------------------------------------------------------+\n"
        "| Rugrats    | Tommy Pickles, Chuckie Finster, Phillip DeVille, Lillian DeVille, Angelica Pickles, |\n"
        "|            | Dil Pickles                                                                         |\n"
        "+------------+-------------------------------------------------------------------------------------+\n"
        "| South Park | Stan Marsh, Kyle Broflovski, Eric Cartman, Kenny McCormick                          |\n"
        "+------------+-------------------------------------------------------------------------------------+"
    )
    assert actual == expected

    # Justify right.
    table.justify_columns = {1: "right"}
    actual = table.table
    expected = (
        "+------------+-------------------------------------------------------------------------------------+\n"
        "| Show       |                                                                          Characters |\n"
        "+------------+-------------------------------------------------------------------------------------+\n"
        "| Rugrats    | Tommy Pickles, Chuckie Finster, Phillip DeVille, Lillian DeVille, Angelica Pickles, |\n"
        "|            |                                                                         Dil Pickles |\n"
        "+------------+-------------------------------------------------------------------------------------+\n"
        "| South Park |                          Stan Marsh, Kyle Broflovski, Eric Cartman, Kenny McCormick |\n"
        "+------------+-------------------------------------------------------------------------------------+"
    )
    assert actual == expected


@pytest.mark.skipif(str(not IS_WINDOWS))
@pytest.mark.skip  # https://github.com/Robpol86/terminaltables3/issues/44
def test_windows_screenshot(tmpdir):
    """Test on Windows in a new console window. Take a screenshot to verify it works.

    :param tmpdir: pytest fixture.
    """
    script = tmpdir.join("script.py")
    command = [sys.executable, str(script)]
    screenshot = PROJECT_ROOT.join("test_ascii_table.png")
    if screenshot.check():
        screenshot.remove()

    # Generate script.
    script_template = dedent(
        """\
    from __future__ import print_function
    import os, time
    from colorclass import Color, Windows
    from terminaltables3 import AsciiTable
    Windows.enable(auto_colors=True)
    stop_after = time.time() + 20

    table_data = [
        [Color('{b}Name{/b}'), Color('{b}Color{/b}'), Color('{b}Misc{/b}')],
        ['Avocado', Color('{autogreen}green{/fg}'), 100],
        ['Tomato', Color('{autored}red{/fg}'), 0.5],
        ['Lettuce', Color('{autogreen}green{/fg}'), None],
    ]
    print(AsciiTable(table_data).table)

    print('Waiting for screenshot_until_match()...')
    while not os.path.exists(r'%s') and time.time() < stop_after:
        time.sleep(0.5)
    """
    )
    script_contents = script_template % str(screenshot)
    script.write(script_contents.encode("utf-8"), mode="wb")

    # Setup expected.
    sub_images = [str(p) for p in HERE.listdir("sub_ascii_*.bmp")]
    assert sub_images

    # Run.
    with RunNewConsole(command) as gen:
        screenshot_until_match(str(screenshot), 15, sub_images, 1, gen)



"""Generate simple tables in terminals from a nested list of strings.

Use SingleTable or DoubleTable instead of AsciiTable for box-drawing characters.

https://github.com/Robpol86/terminaltables3
https://pypi.python.org/pypi/terminaltables3
"""

from terminaltables3.ascii_table import AsciiTable  # noqa
from terminaltables3.github_table import GithubFlavoredMarkdownTable  # noqa
from terminaltables3.other_tables import DoubleTable  # noqa
from terminaltables3.other_tables import PorcelainTable  # noqa
from terminaltables3.other_tables import SingleTable  # noqa

__author__ = "@Robpol86"
__license__ = "MIT"
__version__ = "3.1.0"



"""Test function in module."""

import pytest
from colorama import Fore
from colorclass import Color
from termcolor import colored

from terminaltables3.width_and_alignment import max_dimensions


@pytest.mark.parametrize(
    "table_data,expected_w,expected_h",
    [
        ([[]], [], [0]),
        ([[""]], [0], [0]),
        ([["", ""]], [0, 0], [0]),
        ([[], []], [], [0, 0]),
        ([[""], [""]], [0], [0, 0]),
        ([["", ""], ["", ""]], [0, 0], [0, 0]),
    ],
)
def test_zero_length(table_data, expected_w, expected_h):
    """Test zero-length or empty tables.

    :param list table_data: Input table data to test.
    :param list expected_w: Expected widths.
    :param list expected_h: Expected heights.
    """
    actual = max_dimensions(table_data)
    assert actual == (expected_w, expected_h, expected_w, expected_h)


def test_single_line():
    """Test widths."""
    table_data = [
        ["Name", "Color", "Type"],
        ["Avocado", "green", "nut"],
        ["Tomato", "red", "fruit"],
        ["Lettuce", "green", "vegetable"],
    ]
    assert max_dimensions(table_data, 1, 1) == (
        [7, 5, 9],
        [1, 1, 1, 1],
        [9, 7, 11],
        [1, 1, 1, 1],
    )

    table_data.append(["Watermelon", "green", "fruit"])
    assert max_dimensions(table_data, 2, 2) == (
        [10, 5, 9],
        [1, 1, 1, 1, 1],
        [14, 9, 13],
        [1, 1, 1, 1, 1],
    )


def test_multi_line():
    """Test heights."""
    table_data = [
        ["One\nTwo", "Buckle\nMy\nShoe"],
    ]
    assert max_dimensions(table_data, 0, 0, 1, 1) == ([3, 6], [3], [3, 6], [5])

    table_data = [
        ["Show", "Characters"],
        [
            "Rugrats",
            (
                "Tommy Pickles, Chuckie Finster, Phillip DeVille, Lillian DeVille, Angelica Pickles,\n"
                "Susie Carmichael, Dil Pickles, Kimi Finster, Spike"
            ),
        ],
        ["South Park", "Stan Marsh, Kyle Broflovski, Eric Cartman, Kenny McCormick"],
    ]
    assert max_dimensions(table_data, 0, 0, 2, 2) == (
        [10, 83],
        [1, 2, 1],
        [10, 83],
        [5, 6, 5],
    )


def test_trailing_newline():
    r"""Test with trailing \n."""
    table_data = [
        ["Row One\n<blank>"],
        ["<blank>\nRow Two"],
        ["Row Three\n"],
        ["\nRow Four"],
    ]
    assert max_dimensions(table_data) == ([9], [2, 2, 2, 2], [9], [2, 2, 2, 2])


def test_colors_cjk_rtl():
    """Test color text, CJK characters, and RTL characters."""
    table_data = [
        [Color("{blue}Test{/blue}")],
        [Fore.BLUE + "Test" + Fore.RESET],
        [colored("Test", "blue")],
    ]
    assert max_dimensions(table_data) == ([4], [1, 1, 1], [4], [1, 1, 1])

    table_data = [
        ["蓝色"],
        ["世界你好"],
    ]
    assert max_dimensions(table_data) == ([8], [1, 1], [8], [1, 1])

    table_data = [
        ["שלום"],
        ["معرب"],
    ]
    assert max_dimensions(table_data) == ([4], [1, 1], [4], [1, 1])


def test_non_string():
    """Test with non-string values."""
    table_data = [
        [123, 0.9, None, True, False],
    ]
    assert max_dimensions(table_data) == ([3, 3, 4, 4, 5], [1], [3, 3, 4, 4, 5], [1])



"""Test example scripts."""

import os
import subprocess
import sys

import pytest

from tests import PROJECT_ROOT


@pytest.mark.parametrize("filename", map("example{}.py".format, (1, 2, 3)))
def test(filename):
    """Test with subprocess.

    :param str filename: Example script filename to run.
    """
    command = [sys.executable, str(PROJECT_ROOT.join(filename))]
    env = dict(os.environ, PYTHONIOENCODING="utf-8")

    # Run.
    proc = subprocess.Popen(
        command, env=env, stderr=subprocess.STDOUT, stdout=subprocess.PIPE
    )
    output = proc.communicate()[0]

    # Verify.
    try:
        assert proc.poll() == 0
    except AssertionError:
        print(output)
        raise



"""Test property in BaseTable class."""

from colorama import Fore
from colorclass import Color
from termcolor import colored

from terminaltables3.base_table import BaseTable


def test_ascii():
    """Test with ASCII characters."""
    table_data = [
        ["Name", "Color", "Type"],
        ["Avocado", "green", "nut"],
        ["Tomato", "red", "fruit"],
        ["Lettuce", "green", "vegetable"],
    ]
    table = BaseTable(table_data)
    actual = table.table

    expected = (
        "+---------+-------+-----------+\n"
        "| Name    | Color | Type      |\n"
        "+---------+-------+-----------+\n"
        "| Avocado | green | nut       |\n"
        "| Tomato  | red   | fruit     |\n"
        "| Lettuce | green | vegetable |\n"
        "+---------+-------+-----------+"
    )

    assert actual == expected


def test_int():
    """Test with integers instead of strings."""
    table_data = [
        [100, 10, 1],
        [0, 3, 6],
        [1, 4, 7],
        [2, 5, 8],
    ]
    table = BaseTable(table_data, 1234567890)
    actual = table.table

    expected = (
        "+1234567890+---+\n"
        "| 100 | 10 | 1 |\n"
        "+-----+----+---+\n"
        "| 0   | 3  | 6 |\n"
        "| 1   | 4  | 7 |\n"
        "| 2   | 5  | 8 |\n"
        "+-----+----+---+"
    )

    assert actual == expected


def test_float():
    """Test with floats instead of strings."""
    table_data = [
        [1.0, 22.0, 333.0],
        [0.1, 3.1, 6.1],
        [1.1, 4.1, 7.1],
        [2.1, 5.1, 8.1],
    ]
    table = BaseTable(table_data, 0.12345678)
    actual = table.table

    expected = (
        "+0.12345678--+-------+\n"
        "| 1.0 | 22.0 | 333.0 |\n"
        "+-----+------+-------+\n"
        "| 0.1 | 3.1  | 6.1   |\n"
        "| 1.1 | 4.1  | 7.1   |\n"
        "| 2.1 | 5.1  | 8.1   |\n"
        "+-----+------+-------+"
    )

    assert actual == expected


def test_bool_none():
    """Test with NoneType/boolean instead of strings."""
    table_data = [
        [True, False, None],
        [True, False, None],
        [False, None, True],
        [None, True, False],
    ]
    table = BaseTable(table_data, True)
    actual = table.table

    expected = (
        "+True---+-------+-------+\n"
        "| True  | False | None  |\n"
        "+-------+-------+-------+\n"
        "| True  | False | None  |\n"
        "| False | None  | True  |\n"
        "| None  | True  | False |\n"
        "+-------+-------+-------+"
    )

    assert actual == expected


def test_cjk():
    """Test with CJK characters."""
    table_data = [
        ["CJK"],
        ["蓝色"],
        ["世界你好"],
    ]
    table = BaseTable(table_data)
    actual = table.table

    expected = (
        "+----------+\n"
        "| CJK      |\n"
        "+----------+\n"
        "| 蓝色     |\n"
        "| 世界你好 |\n"
        "+----------+"
    )

    assert actual == expected


def test_rtl():
    """Test with RTL characters."""
    table_data = [
        ["RTL"],
        ["שלום"],
        ["معرب"],
    ]
    table = BaseTable(table_data)
    actual = table.table

    expected = (
        "+------+\n" "| RTL  |\n" "+------+\n" "| שלום |\n" "| معرب |\n" "+------+"
    )

    assert actual == expected


def test_rtl_large():
    """Test large table of RTL characters."""
    table_data = [
        ["اكتب", "اللون", "اسم"],
        ["البندق", "أخضر", "أفوكادو"],
        ["ثمرة", "أحمر", "بندورة"],
        ["الخضروات", "أخضر", "الخس"],
    ]
    table = BaseTable(table_data, "جوجل المترجم")
    actual = table.table

    expected = (
        "+جوجل المترجم------+---------+\n"
        "| اكتب     | اللون | اسم     |\n"
        "+----------+-------+---------+\n"
        "| البندق   | أخضر  | أفوكادو |\n"
        "| ثمرة     | أحمر  | بندورة  |\n"
        "| الخضروات | أخضر  | الخس    |\n"
        "+----------+-------+---------+"
    )

    assert actual == expected


def test_color():
    """Test with color characters."""
    table_data = [
        [
            "ansi",
            "\033[31mRed\033[39m",
            "\033[32mGreen\033[39m",
            "\033[34mBlue\033[39m",
        ],
        [
            "colorclass",
            Color("{red}Red{/red}"),
            Color("{green}Green{/green}"),
            Color("{blue}Blue{/blue}"),
        ],
        [
            "colorama",
            Fore.RED + "Red" + Fore.RESET,
            Fore.GREEN + "Green" + Fore.RESET,
            Fore.BLUE + "Blue" + Fore.RESET,
        ],
        [
            "termcolor",
            colored("Red", "red"),
            colored("Green", "green"),
            colored("Blue", "blue"),
        ],
    ]
    table = BaseTable(table_data)
    table.inner_heading_row_border = False
    actual = table.table

    expected = (
        "+------------+-----+-------+------+\n"
        "| ansi       | \033[31mRed\033[39m | \033[32mGreen\033[39m | \033[34mBlue\033[39m |\n"
        "| colorclass | \033[31mRed\033[39m | \033[32mGreen\033[39m | \033[34mBlue\033[39m |\n"
        "| colorama   | \033[31mRed\033[39m | \033[32mGreen\033[39m | \033[34mBlue\033[39m |\n"
        "| termcolor  | \033[31mRed\033[0m | \033[32mGreen\033[0m | \033[34mBlue\033[0m |\n"
        "+------------+-----+-------+------+"
    )

    assert actual == expected



"""Test function in module."""

import pytest

from terminaltables3.terminal_io import (
    DEFAULT_HEIGHT,
    DEFAULT_WIDTH,
    INVALID_HANDLE_VALUE,
    IS_WINDOWS,
    terminal_size,
)

from tests.test_terminal_io import MockKernel32


@pytest.mark.parametrize("stderr", [1, INVALID_HANDLE_VALUE])
@pytest.mark.parametrize("stdout", [2, INVALID_HANDLE_VALUE])
def test_windows(monkeypatch, stderr, stdout):
    """Test function with IS_WINDOWS=True.

    :param monkeypatch: pytest fixture.
    :param int stderr: Mock handle value.
    :param int stdout: Mock handle value.
    """
    monkeypatch.setattr("terminaltables3.terminal_io.IS_WINDOWS", True)

    kernel32 = MockKernel32(stderr=stderr, stdout=stdout)
    width, height = terminal_size(kernel32)

    if stderr == INVALID_HANDLE_VALUE and stdout == INVALID_HANDLE_VALUE:
        assert width == DEFAULT_WIDTH
        assert height == DEFAULT_HEIGHT
    elif stdout == INVALID_HANDLE_VALUE:
        assert width == 119
        assert height == 29
    elif stderr == INVALID_HANDLE_VALUE:
        assert width == 75
        assert height == 28
    else:
        assert width == 119
        assert height == 29


@pytest.mark.skipif(str(IS_WINDOWS))
def test_nix(monkeypatch):
    """Test function with IS_WINDOWS=False.

    :param monkeypatch: pytest fixture.
    """
    # Test exception (no terminal within pytest).
    width, height = terminal_size()
    assert width == DEFAULT_WIDTH
    assert height == DEFAULT_HEIGHT

    # Test mocked.
    monkeypatch.setattr("fcntl.ioctl", lambda *_: b"\x1d\x00w\x00\xca\x02\x96\x01")
    width, height = terminal_size()
    assert width == 119
    assert height == 29



"""Test function in module."""

import pytest
from colorama import Fore, Style
from colorclass import Color
from termcolor import colored

from terminaltables3.build import build_border


@pytest.mark.parametrize(
    "outer_widths,horizontal,left,intersect,right,expected",
    [
        ([5, 6, 7], "-", "<", "+", ">", "<-----+------+------->"),
        ([1, 1, 1], "-", "", "", "", "---"),
        ([1, 1, 1], "", "", "", "", ""),
        ([1], "-", "<", "+", ">", "<->"),
        ([], "-", "<", "+", ">", "<>"),
    ],
)
def test_no_title(outer_widths, horizontal, left, intersect, right, expected):
    """Test without title.

    :param iter outer_widths: List of integers representing column widths with padding.
    :param str horizontal: Character to stretch across each column.
    :param str left: Left border.
    :param str intersect: Column separator.
    :param str right: Right border.
    :param str expected: Expected output.
    """
    actual = build_border(outer_widths, horizontal, left, intersect, right)
    assert "".join(actual) == expected


@pytest.mark.parametrize(
    "outer_widths,intersect,expected",
    [
        ([20], "+", "Applications--------"),
        ([20], "", "Applications--------"),
        ([15, 5], "+", "Applications---+-----"),
        ([15, 5], "", "Applications--------"),
        ([12], "+", "Applications"),
        ([12], "", "Applications"),
        ([12, 1], "+", "Applications+-"),
        ([12, 1], "", "Applications-"),
        ([12, 0], "+", "Applications+"),
        ([12, 0], "", "Applications"),
    ],
)
@pytest.mark.parametrize("left,right", [("", ""), ("<", ">")])
def test_first_column_fit(outer_widths, left, intersect, right, expected):
    """Test with title that fits in the first column.

    :param iter outer_widths: List of integers representing column widths with padding.
    :param str left: Left border.
    :param str intersect: Column separator.
    :param str right: Right border.
    :param str expected: Expected output.
    """
    if left and right:
        expected = left + expected + right
    actual = build_border(
        outer_widths, "-", left, intersect, right, title="Applications"
    )
    assert "".join(actual) == expected


@pytest.mark.parametrize(
    "outer_widths,expected",
    [
        ([20], "Applications--------"),
        ([10, 10], "Applications--------"),
        ([5, 5, 5, 5], "Applications--------"),
        ([3, 2, 3, 2, 3, 2, 3, 2], "Applications--------"),
        ([1] * 20, "Applications--------"),
        ([10, 5], "Applications---"),
        ([9, 5], "Applications--"),
        ([8, 5], "Applications-"),
        ([7, 5], "Applications"),
        ([6, 5], "-----------"),
    ],
)
@pytest.mark.parametrize("left,right", [("", ""), ("<", ">")])
def test_no_intersect(outer_widths, left, right, expected):
    """Test with no column dividers.

    :param iter outer_widths: List of integers representing column widths.
    :param str left: Left border.
    :param str right: Right border.
    :param str expected: Expected output.
    """
    if left and right:
        expected = left + expected + right
    actual = build_border(outer_widths, "-", left, "", right, title="Applications")
    assert "".join(actual) == expected


@pytest.mark.parametrize(
    "outer_widths,expected",
    [
        ([20], "Applications--------"),
        ([0, 20], "Applications---------"),
        ([20, 0], "Applications--------+"),
        ([0, 0, 20], "Applications----------"),
        ([20, 0, 0], "Applications--------++"),
        ([10, 10], "Applications---------"),
        ([11, 9], "Applications---------"),
        ([12, 8], "Applications+--------"),
        ([13, 7], "Applications-+-------"),
        ([5, 5, 5, 5], "Applications-----+-----"),
        ([4, 4, 6, 6], "Applications----+------"),
        ([3, 3, 7, 7], "Applications---+-------"),
        ([2, 2, 7, 9], "Applications-+---------"),
        ([1, 1, 9, 9], "Applications-+---------"),
        ([2, 2, 2, 2, 2, 2, 2], "Applications--+--+--"),
        ([1, 1, 1, 1, 1, 1, 1, 1, 1, 1], "Applications-+-+-+-"),
        (
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            "Applications++++++++",
        ),
        ([2, 2, 2, 2], "--+--+--+--"),
        ([1, 1, 1, 1, 1], "-+-+-+-+-"),
        ([0, 0, 0, 0, 0, 0, 0, 0, 0, 0], "+++++++++"),
    ],
)
@pytest.mark.parametrize("left,right", [("", ""), ("<", ">")])
def test_intersect(outer_widths, left, right, expected):
    """Test with column dividers.

    :param iter outer_widths: List of integers representing column widths.
    :param str left: Left border.
    :param str right: Right border.
    :param str expected: Expected output.
    """
    if left and right:
        expected = left + expected + right
    actual = build_border(outer_widths, "-", left, "+", right, title="Applications")
    assert "".join(actual) == expected


@pytest.mark.parametrize(
    "outer_widths,intersect,expected",
    [
        ([12], "+", "蓝色--------"),
        ([12], "", "蓝色--------"),
        ([7, 5], "+", "蓝色---+-----"),
        ([7, 5], "", "蓝色--------"),
        ([4], "+", "蓝色"),
        ([4], "", "蓝色"),
        ([4, 1], "+", "蓝色+-"),
        ([4, 1], "", "蓝色-"),
        ([4, 0], "+", "蓝色+"),
        ([4, 0], "", "蓝色"),
        ([12], "", "蓝色--------"),
        ([6, 6], "", "蓝色--------"),
        ([3, 3, 3, 3], "", "蓝色--------"),
        ([2, 1, 2, 1, 2, 1, 2, 1], "", "蓝色--------"),
        ([1] * 12, "", "蓝色--------"),
        ([2, 4], "", "蓝色--"),
        ([1, 4], "", "蓝色-"),
        ([1, 3], "", "蓝色"),
        ([1, 2], "", "---"),
        ([2], "", "--"),
        ([12], "+", "蓝色--------"),
        ([0, 12], "+", "蓝色---------"),
        ([12, 0], "+", "蓝色--------+"),
        ([0, 0, 12], "+", "蓝色----------"),
        ([12, 0, 0], "+", "蓝色--------++"),
        ([3, 3], "+", "蓝色---"),
        ([4, 2], "+", "蓝色+--"),
        ([5, 1], "+", "蓝色-+-"),
        ([3, 3, 3, 3], "+", "蓝色---+---+---"),
        ([2, 2, 4, 4], "+", "蓝色-+----+----"),
        ([1, 1, 5, 5], "+", "蓝色-----+-----"),
        ([2, 2, 2, 2], "+", "蓝色-+--+--"),
        ([1, 1, 1, 1, 1], "+", "蓝色-+-+-"),
        ([0, 0, 0, 0, 0, 0, 0], "+", "蓝色++"),
        ([1, 1], "+", "-+-"),
    ],
)
@pytest.mark.parametrize("left,right", [("", ""), ("<", ">")])
def test_cjk(outer_widths, left, intersect, right, expected):
    """Test with CJK characters in title.

    :param iter outer_widths: List of integers representing column widths.
    :param str left: Left border.
    :param str intersect: Column separator.
    :param str right: Right border.
    :param str expected: Expected output.
    """
    if left and right:
        expected = left + expected + right
    actual = build_border(outer_widths, "-", left, intersect, right, title="蓝色")
    assert "".join(actual) == expected


@pytest.mark.parametrize(
    "outer_widths,intersect,expected",
    [
        ([12], "+", "معرب--------"),
        ([12], "", "معرب--------"),
        ([7, 5], "+", "معرب---+-----"),
        ([7, 5], "", "معرب--------"),
        ([4], "+", "معرب"),
        ([4], "", "معرب"),
        ([4, 1], "+", "معرب+-"),
        ([4, 1], "", "معرب-"),
        ([4, 0], "+", "معرب+"),
        ([4, 0], "", "معرب"),
        ([12], "", "معرب--------"),
        ([6, 6], "", "معرب--------"),
        ([3, 3, 3, 3], "", "معرب--------"),
        ([2, 1, 2, 1, 2, 1, 2, 1], "", "معرب--------"),
        ([1] * 12, "", "معرب--------"),
        ([2, 4], "", "معرب--"),
        ([1, 4], "", "معرب-"),
        ([1, 3], "", "معرب"),
        ([1, 2], "", "---"),
        ([2], "", "--"),
        ([12], "+", "معرب--------"),
        ([0, 12], "+", "معرب---------"),
        ([12, 0], "+", "معرب--------+"),
        ([0, 0, 12], "+", "معرب----------"),
        ([12, 0, 0], "+", "معرب--------++"),
        ([3, 3], "+", "معرب---"),
        ([4, 2], "+", "معرب+--"),
        ([5, 1], "+", "معرب-+-"),
        ([3, 3, 3, 3], "+", "معرب---+---+---"),
        ([2, 2, 4, 4], "+", "معرب-+----+----"),
        ([1, 1, 5, 5], "+", "معرب-----+-----"),
        ([2, 2, 2, 2], "+", "معرب-+--+--"),
        ([1, 1, 1, 1, 1], "+", "معرب-+-+-"),
        ([0, 0, 0, 0, 0, 0, 0], "+", "معرب++"),
        ([1, 1], "+", "-+-"),
    ],
)
@pytest.mark.parametrize("left,right", [("", ""), ("<", ">")])
def test_rtl(outer_widths, left, intersect, right, expected):
    """Test with RTL characters in title.

    :param iter outer_widths: List of integers representing column widths.
    :param str left: Left border.
    :param str intersect: Column separator.
    :param str right: Right border.
    :param str expected: Expected output.
    """
    if left and right:
        expected = left + expected + right
    actual = build_border(outer_widths, "-", left, intersect, right, title="معرب")
    assert "".join(actual) == expected


@pytest.mark.parametrize(
    "outer_widths,intersect,expected",
    [
        ([12], "+", "\x1b[34mTEST\x1b[0m--------"),
        ([12], "", "\x1b[34mTEST\x1b[0m--------"),
        ([7, 5], "+", "\x1b[34mTEST\x1b[0m---+-----"),
        ([7, 5], "", "\x1b[34mTEST\x1b[0m--------"),
        ([4], "+", "\x1b[34mTEST\x1b[0m"),
        ([4], "", "\x1b[34mTEST\x1b[0m"),
        ([4, 1], "+", "\x1b[34mTEST\x1b[0m+-"),
        ([4, 1], "", "\x1b[34mTEST\x1b[0m-"),
        ([4, 0], "+", "\x1b[34mTEST\x1b[0m+"),
        ([4, 0], "", "\x1b[34mTEST\x1b[0m"),
        ([12], "", "\x1b[34mTEST\x1b[0m--------"),
        ([6, 6], "", "\x1b[34mTEST\x1b[0m--------"),
        ([3, 3, 3, 3], "", "\x1b[34mTEST\x1b[0m--------"),
        ([2, 1, 2, 1, 2, 1, 2, 1], "", "\x1b[34mTEST\x1b[0m--------"),
        ([1] * 12, "", "\x1b[34mTEST\x1b[0m--------"),
        ([2, 4], "", "\x1b[34mTEST\x1b[0m--"),
        ([1, 4], "", "\x1b[34mTEST\x1b[0m-"),
        ([1, 3], "", "\x1b[34mTEST\x1b[0m"),
        ([1, 2], "", "---"),
        ([12], "+", "\x1b[34mTEST\x1b[0m--------"),
        ([0, 12], "+", "\x1b[34mTEST\x1b[0m---------"),
        ([12, 0], "+", "\x1b[34mTEST\x1b[0m--------+"),
        ([0, 0, 12], "+", "\x1b[34mTEST\x1b[0m----------"),
        ([12, 0, 0], "+", "\x1b[34mTEST\x1b[0m--------++"),
        ([3, 3], "+", "\x1b[34mTEST\x1b[0m---"),
        ([4, 2], "+", "\x1b[34mTEST\x1b[0m+--"),
        ([5, 1], "+", "\x1b[34mTEST\x1b[0m-+-"),
        ([3, 3, 3, 3], "+", "\x1b[34mTEST\x1b[0m---+---+---"),
        ([2, 2, 4, 4], "+", "\x1b[34mTEST\x1b[0m-+----+----"),
        ([1, 1, 5, 5], "+", "\x1b[34mTEST\x1b[0m-----+-----"),
        ([2, 2, 2, 2], "+", "\x1b[34mTEST\x1b[0m-+--+--"),
        ([1, 1, 1, 1, 1], "+", "\x1b[34mTEST\x1b[0m-+-+-"),
        ([0, 0, 0, 0, 0, 0, 0], "+", "\x1b[34mTEST\x1b[0m++"),
        ([1, 1], "+", "-+-"),
    ],
)
@pytest.mark.parametrize("left,right", [("", ""), ("<", ">")])
@pytest.mark.parametrize(
    "title",
    [
        "\x1b[34mTEST\x1b[0m",
        Color("{blue}TEST{/all}"),
        Fore.BLUE + "TEST" + Style.RESET_ALL,
        colored("TEST", "blue"),
    ],
)
def test_colors(outer_widths, left, intersect, right, title, expected):
    """Test with color title characters.

    :param iter outer_widths: List of integers representing column widths with padding.
    :param str left: Left border.
    :param str intersect: Column separator.
    :param str right: Right border.
    :param title: Title in border with color codes.
    :param str expected: Expected output.
    """
    if left and right:
        expected = left + expected + right
    actual = build_border(outer_widths, "-", left, intersect, right, title=title)
    assert "".join(actual) == expected


@pytest.mark.parametrize(
    "outer_widths,title,expected",
    [
        ([3, 3, 3], 123, "<123+---+--->"),
        ([3, 3, 3], 0.9, "<0.9+---+--->"),
        ([3, 3, 3], True, "<True---+--->"),
        ([3, 3, 3], False, "<False--+--->"),
    ],
)
def test_non_string(outer_widths, title, expected):
    """Test with non-string values.

    :param iter outer_widths: List of integers representing column widths with padding.
    :param title: Title in border.
    :param str expected: Expected output.
    """
    actual = build_border(outer_widths, "-", "<", "+", ">", title=title)
    assert "".join(actual) == expected



"""Additional simple tables defined here."""

from terminaltables3.ascii_table import AsciiTable
from terminaltables3.terminal_io import IS_WINDOWS


class UnixTable(AsciiTable):
    """Draw a table using box-drawing characters on Unix platforms. Table borders won't have any gaps between lines.

    Similar to the tables shown on PC BIOS boot messages, but not double-lined.
    """

    CHAR_F_INNER_HORIZONTAL = "\033(0\x71\033(B"
    CHAR_F_INNER_INTERSECT = "\033(0\x6e\033(B"
    CHAR_F_INNER_VERTICAL = "\033(0\x78\033(B"
    CHAR_F_OUTER_LEFT_INTERSECT = "\033(0\x74\033(B"
    CHAR_F_OUTER_LEFT_VERTICAL = "\033(0\x78\033(B"
    CHAR_F_OUTER_RIGHT_INTERSECT = "\033(0\x75\033(B"
    CHAR_F_OUTER_RIGHT_VERTICAL = "\033(0\x78\033(B"
    CHAR_H_INNER_HORIZONTAL = "\033(0\x71\033(B"
    CHAR_H_INNER_INTERSECT = "\033(0\x6e\033(B"
    CHAR_H_INNER_VERTICAL = "\033(0\x78\033(B"
    CHAR_H_OUTER_LEFT_INTERSECT = "\033(0\x74\033(B"
    CHAR_H_OUTER_LEFT_VERTICAL = "\033(0\x78\033(B"
    CHAR_H_OUTER_RIGHT_INTERSECT = "\033(0\x75\033(B"
    CHAR_H_OUTER_RIGHT_VERTICAL = "\033(0\x78\033(B"
    CHAR_INNER_HORIZONTAL = "\033(0\x71\033(B"
    CHAR_INNER_INTERSECT = "\033(0\x6e\033(B"
    CHAR_INNER_VERTICAL = "\033(0\x78\033(B"
    CHAR_OUTER_BOTTOM_HORIZONTAL = "\033(0\x71\033(B"
    CHAR_OUTER_BOTTOM_INTERSECT = "\033(0\x76\033(B"
    CHAR_OUTER_BOTTOM_LEFT = "\033(0\x6d\033(B"
    CHAR_OUTER_BOTTOM_RIGHT = "\033(0\x6a\033(B"
    CHAR_OUTER_LEFT_INTERSECT = "\033(0\x74\033(B"
    CHAR_OUTER_LEFT_VERTICAL = "\033(0\x78\033(B"
    CHAR_OUTER_RIGHT_INTERSECT = "\033(0\x75\033(B"
    CHAR_OUTER_RIGHT_VERTICAL = "\033(0\x78\033(B"
    CHAR_OUTER_TOP_HORIZONTAL = "\033(0\x71\033(B"
    CHAR_OUTER_TOP_INTERSECT = "\033(0\x77\033(B"
    CHAR_OUTER_TOP_LEFT = "\033(0\x6c\033(B"
    CHAR_OUTER_TOP_RIGHT = "\033(0\x6b\033(B"

    @property
    def table(self):
        """Return a large string of the entire table ready to be printed to the terminal."""
        ascii_table = super().table
        optimized = ascii_table.replace("\033(B\033(0", "")
        return optimized


class WindowsTable(AsciiTable):
    """Draw a table using box-drawing characters on Windows platforms. This uses Code Page 437. Single-line borders.

    From: http://en.wikipedia.org/wiki/Code_page_437#Characters
    """

    CHAR_F_INNER_HORIZONTAL = b"\xc4".decode("ibm437")
    CHAR_F_INNER_INTERSECT = b"\xc5".decode("ibm437")
    CHAR_F_INNER_VERTICAL = b"\xb3".decode("ibm437")
    CHAR_F_OUTER_LEFT_INTERSECT = b"\xc3".decode("ibm437")
    CHAR_F_OUTER_LEFT_VERTICAL = b"\xb3".decode("ibm437")
    CHAR_F_OUTER_RIGHT_INTERSECT = b"\xb4".decode("ibm437")
    CHAR_F_OUTER_RIGHT_VERTICAL = b"\xb3".decode("ibm437")
    CHAR_H_INNER_HORIZONTAL = b"\xc4".decode("ibm437")
    CHAR_H_INNER_INTERSECT = b"\xc5".decode("ibm437")
    CHAR_H_INNER_VERTICAL = b"\xb3".decode("ibm437")
    CHAR_H_OUTER_LEFT_INTERSECT = b"\xc3".decode("ibm437")
    CHAR_H_OUTER_LEFT_VERTICAL = b"\xb3".decode("ibm437")
    CHAR_H_OUTER_RIGHT_INTERSECT = b"\xb4".decode("ibm437")
    CHAR_H_OUTER_RIGHT_VERTICAL = b"\xb3".decode("ibm437")
    CHAR_INNER_HORIZONTAL = b"\xc4".decode("ibm437")
    CHAR_INNER_INTERSECT = b"\xc5".decode("ibm437")
    CHAR_INNER_VERTICAL = b"\xb3".decode("ibm437")
    CHAR_OUTER_BOTTOM_HORIZONTAL = b"\xc4".decode("ibm437")
    CHAR_OUTER_BOTTOM_INTERSECT = b"\xc1".decode("ibm437")
    CHAR_OUTER_BOTTOM_LEFT = b"\xc0".decode("ibm437")
    CHAR_OUTER_BOTTOM_RIGHT = b"\xd9".decode("ibm437")
    CHAR_OUTER_LEFT_INTERSECT = b"\xc3".decode("ibm437")
    CHAR_OUTER_LEFT_VERTICAL = b"\xb3".decode("ibm437")
    CHAR_OUTER_RIGHT_INTERSECT = b"\xb4".decode("ibm437")
    CHAR_OUTER_RIGHT_VERTICAL = b"\xb3".decode("ibm437")
    CHAR_OUTER_TOP_HORIZONTAL = b"\xc4".decode("ibm437")
    CHAR_OUTER_TOP_INTERSECT = b"\xc2".decode("ibm437")
    CHAR_OUTER_TOP_LEFT = b"\xda".decode("ibm437")
    CHAR_OUTER_TOP_RIGHT = b"\xbf".decode("ibm437")


class WindowsTableDouble(AsciiTable):
    """Draw a table using box-drawing characters on Windows platforms. This uses Code Page 437. Double-line borders."""

    CHAR_F_INNER_HORIZONTAL = b"\xcd".decode("ibm437")
    CHAR_F_INNER_INTERSECT = b"\xce".decode("ibm437")
    CHAR_F_INNER_VERTICAL = b"\xba".decode("ibm437")
    CHAR_F_OUTER_LEFT_INTERSECT = b"\xcc".decode("ibm437")
    CHAR_F_OUTER_LEFT_VERTICAL = b"\xba".decode("ibm437")
    CHAR_F_OUTER_RIGHT_INTERSECT = b"\xb9".decode("ibm437")
    CHAR_F_OUTER_RIGHT_VERTICAL = b"\xba".decode("ibm437")
    CHAR_H_INNER_HORIZONTAL = b"\xcd".decode("ibm437")
    CHAR_H_INNER_INTERSECT = b"\xce".decode("ibm437")
    CHAR_H_INNER_VERTICAL = b"\xba".decode("ibm437")
    CHAR_H_OUTER_LEFT_INTERSECT = b"\xcc".decode("ibm437")
    CHAR_H_OUTER_LEFT_VERTICAL = b"\xba".decode("ibm437")
    CHAR_H_OUTER_RIGHT_INTERSECT = b"\xb9".decode("ibm437")
    CHAR_H_OUTER_RIGHT_VERTICAL = b"\xba".decode("ibm437")
    CHAR_INNER_HORIZONTAL = b"\xcd".decode("ibm437")
    CHAR_INNER_INTERSECT = b"\xce".decode("ibm437")
    CHAR_INNER_VERTICAL = b"\xba".decode("ibm437")
    CHAR_OUTER_BOTTOM_HORIZONTAL = b"\xcd".decode("ibm437")
    CHAR_OUTER_BOTTOM_INTERSECT = b"\xca".decode("ibm437")
    CHAR_OUTER_BOTTOM_LEFT = b"\xc8".decode("ibm437")
    CHAR_OUTER_BOTTOM_RIGHT = b"\xbc".decode("ibm437")
    CHAR_OUTER_LEFT_INTERSECT = b"\xcc".decode("ibm437")
    CHAR_OUTER_LEFT_VERTICAL = b"\xba".decode("ibm437")
    CHAR_OUTER_RIGHT_INTERSECT = b"\xb9".decode("ibm437")
    CHAR_OUTER_RIGHT_VERTICAL = b"\xba".decode("ibm437")
    CHAR_OUTER_TOP_HORIZONTAL = b"\xcd".decode("ibm437")
    CHAR_OUTER_TOP_INTERSECT = b"\xcb".decode("ibm437")
    CHAR_OUTER_TOP_LEFT = b"\xc9".decode("ibm437")
    CHAR_OUTER_TOP_RIGHT = b"\xbb".decode("ibm437")


class SingleTable(WindowsTable if IS_WINDOWS else UnixTable):
    """Cross-platform table with single-line box-drawing characters.

    :ivar iter table_data: List (empty or list of lists of strings) representing the table.
    :ivar str title: Optional title to show within the top border of the table.
    :ivar bool inner_column_border: Separates columns.
    :ivar bool inner_footing_row_border: Show a border before the last row.
    :ivar bool inner_heading_row_border: Show a border after the first row.
    :ivar bool inner_row_border: Show a border in between every row.
    :ivar bool outer_border: Show the top, left, right, and bottom border.
    :ivar dict justify_columns: Horizontal justification. Keys are column indexes (int). Values are right/left/center.
    :ivar int padding_left: Number of spaces to pad on the left side of every cell.
    :ivar int padding_right: Number of spaces to pad on the right side of every cell.
    """


class DoubleTable(WindowsTableDouble):
    """Cross-platform table with box-drawing characters. On Windows it's double borders, on Linux/OSX it's unicode.

    :ivar iter table_data: List (empty or list of lists of strings) representing the table.
    :ivar str title: Optional title to show within the top border of the table.
    :ivar bool inner_column_border: Separates columns.
    :ivar bool inner_footing_row_border: Show a border before the last row.
    :ivar bool inner_heading_row_border: Show a border after the first row.
    :ivar bool inner_row_border: Show a border in between every row.
    :ivar bool outer_border: Show the top, left, right, and bottom border.
    :ivar dict justify_columns: Horizontal justification. Keys are column indexes (int). Values are right/left/center.
    :ivar int padding_left: Number of spaces to pad on the left side of every cell.
    :ivar int padding_right: Number of spaces to pad on the right side of every cell.
    """


class PorcelainTable(AsciiTable):
    """An AsciiTable stripped to a minimum.

    Meant to be machine passable and roughly follow format set by git --porcelain option (hence the name).

    :ivar iter table_data: List (empty or list of lists of strings) representing the table.
    """

    def __init__(self, table_data):
        """Constructor.

        :param iter table_data: List (empty or list of lists of strings) representing the table.
        """
        # Porcelain table won't support title since it has no outer birders.
        super().__init__(table_data)

        # Removes outer border, and inner footing and header row borders.
        self.inner_footing_row_border = False
        self.inner_heading_row_border = False
        self.outer_border = False



"""Test method in BaseTable class."""

import pytest

from terminaltables3.base_table import BaseTable


@pytest.mark.parametrize("style", ["heading", "footing", "row"])
def test_single_line(style):
    """Test with single-line row.

    :param str style: Passed to method.
    """
    row = ["Row One Column One", "Two", "Three"]
    table = BaseTable([row])
    actual = [tuple(i) for i in table.gen_row_lines(row, style, [18, 3, 5], 1)]
    expected = [
        ("|", " Row One Column One ", "|", " Two ", "|", " Three ", "|"),
    ]
    assert actual == expected


@pytest.mark.parametrize("style", ["heading", "footing", "row"])
def test_multi_line(style):
    """Test with multi-line row.

    :param str style: Passed to method.
    """
    row = ["Row One\nColumn One", "Two", "Three"]
    table = BaseTable([row])
    actual = [tuple(i) for i in table.gen_row_lines(row, style, [10, 3, 5], 2)]
    expected = [
        ("|", " Row One    ", "|", " Two ", "|", " Three ", "|"),
        ("|", " Column One ", "|", "     ", "|", "       ", "|"),
    ]
    assert actual == expected


@pytest.mark.parametrize("style", ["heading", "footing", "row"])
def test_no_padding_no_borders(style):
    """Test without padding or borders.

    :param str style: Passed to method.
    """
    row = ["Row One\nColumn One", "Two", "Three"]
    table = BaseTable([row])
    table.inner_column_border = False
    table.outer_border = False
    table.padding_left = 0
    table.padding_right = 0
    actual = [tuple(i) for i in table.gen_row_lines(row, style, [10, 3, 5], 2)]
    expected = [
        ("Row One   ", "Two", "Three"),
        ("Column One", "   ", "     "),
    ]
    assert actual == expected


@pytest.mark.parametrize("style", ["heading", "footing", "row"])
def test_uneven(style):
    """Test with row missing cells.

    :param str style: Passed to method.
    """
    row = ["Row One Column One"]
    table = BaseTable([row])
    actual = [tuple(i) for i in table.gen_row_lines(row, style, [18, 3, 5], 1)]
    expected = [
        ("|", " Row One Column One ", "|", "     ", "|", "       ", "|"),
    ]
    assert actual == expected


@pytest.mark.parametrize("style", ["heading", "footing", "row"])
def test_empty_table(style):
    """Test empty table.

    :param str style: Passed to method.
    """
    row = []
    table = BaseTable([row])
    actual = [tuple(i) for i in table.gen_row_lines(row, style, [], 0)]
    expected = [
        ("|", "|"),
    ]
    assert actual == expected



"""AsciiTable is the main table class. To be inherited by other tables. Define convenience methods here."""

from terminaltables3.base_table import BaseTable
from terminaltables3.terminal_io import terminal_size
from terminaltables3.width_and_alignment import (
    column_max_width,
    max_dimensions,
    table_width,
)


class AsciiTable(BaseTable):
    """Draw a table using regular ASCII characters, such as ``+``, ``|``, and ``-``.

    :ivar iter table_data: List (empty or list of lists of strings) representing the table.
    :ivar str title: Optional title to show within the top border of the table.
    :ivar bool inner_column_border: Separates columns.
    :ivar bool inner_footing_row_border: Show a border before the last row.
    :ivar bool inner_heading_row_border: Show a border after the first row.
    :ivar bool inner_row_border: Show a border in between every row.
    :ivar bool outer_border: Show the top, left, right, and bottom border.
    :ivar dict justify_columns: Horizontal justification. Keys are column indexes (int). Values are right/left/center.
    :ivar int padding_left: Number of spaces to pad on the left side of every cell.
    :ivar int padding_right: Number of spaces to pad on the right side of every cell.
    """

    def column_max_width(self, column_number: int) -> int:
        """Return the maximum width of a column based on the current terminal width.

        :param int column_number: The column number to query.

        :return: The max width of the column.
        :rtype: int
        """
        inner_widths = max_dimensions(self.table_data)[0]
        outer_border = 2 if self.outer_border else 0
        inner_border = 1 if self.inner_column_border else 0
        padding = self.padding_left + self.padding_right
        return column_max_width(
            inner_widths, column_number, outer_border, inner_border, padding
        )

    @property
    def column_widths(self) -> list[int]:
        """Return a list of integers representing the widths of each table column without padding."""
        if not self.table_data:
            return []
        return max_dimensions(self.table_data)[0]

    @property
    def ok(self) -> bool:  # Too late to change API. # pylint: disable=invalid-name
        """Return True if the table fits within the terminal width, False if the table breaks."""
        return self.table_width <= terminal_size()[0]

    @property
    def table_width(self) -> int:
        """Return the width of the table including padding and borders."""
        outer_widths = max_dimensions(
            self.table_data, self.padding_left, self.padding_right
        )[2]
        outer_border = 2 if self.outer_border else 0
        inner_border = 1 if self.inner_column_border else 0
        return table_width(outer_widths, outer_border, inner_border)



"""Test function in module."""

import ctypes

import pytest

from terminaltables3.terminal_io import (
    get_console_info,
    INVALID_HANDLE_VALUE,
    IS_WINDOWS,
)

from tests.test_terminal_io import MockKernel32


def test():
    """Test function."""
    # Test live WinError.
    if IS_WINDOWS:
        with pytest.raises(OSError):
            get_console_info(ctypes.windll.kernel32, 0)

    # Test INVALID_HANDLE_VALUE.
    kernel32 = MockKernel32(stderr=1)
    with pytest.raises(OSError):
        get_console_info(kernel32, INVALID_HANDLE_VALUE)

    # Test no error with mock methods.
    width, height = get_console_info(kernel32, 1)
    assert width == 119
    assert height == 29



"""Functions that handle alignment, padding, widths, etc."""

import re
import unicodedata
from typing import Sequence, Tuple

from terminaltables3.terminal_io import terminal_size

RE_COLOR_ANSI = re.compile(r"(\033\[[\d;]+m)")


def visible_width(string: str) -> int:
    """Get the visible width of a unicode string.

    Some CJK unicode characters are more than one byte unlike ASCII and latin unicode characters.

    From: https://github.com/Robpol86/terminaltables3/pull/9

    :param str string: String to measure.

    :return: String's width.
    :rtype: int
    """
    if "\033" in string:
        string = RE_COLOR_ANSI.sub("", string)

    # Convert to unicode.
    try:
        string = string.decode("u8")
    except (AttributeError, UnicodeEncodeError):
        pass

    width = 0
    for char in string:
        if unicodedata.east_asian_width(char) in ("F", "W"):
            width += 2
        else:
            width += 1

    return width


def align_and_pad_cell(
    string: str,
    align: Tuple,
    inner_dimensions: Tuple,
    padding: Sequence[int],
    space: str = " ",
) -> list[str]:
    """Align a string horizontally and vertically. Also add additional padding in both dimensions.

    :param str string: Input string to operate on.
    :param tuple align: Tuple that contains one of left/center/right and/or top/middle/bottom.
    :param tuple inner_dimensions: Width and height ints to expand string to without padding.
    :param iter padding: Number of space chars for left, right, top, and bottom (4 ints).
    :param str space: Character to use as white space for resizing/padding (use single visible chars only).

    :return: Padded cell split into lines.
    :rtype: list
    """
    if not hasattr(string, "splitlines"):
        string = str(string)

    # Handle trailing newlines or empty strings, str.splitlines() does not satisfy.
    lines = string.splitlines() or [""]
    if string.endswith("\n"):
        lines.append("")

    # Vertically align and pad.
    if "bottom" in align:
        lines = (
            ([""] * (inner_dimensions[1] - len(lines) + padding[2]))
            + lines
            + ([""] * padding[3])
        )
    elif "middle" in align:
        delta = inner_dimensions[1] - len(lines)
        lines = (
            ([""] * (delta // 2 + delta % 2 + padding[2]))
            + lines
            + ([""] * (delta // 2 + padding[3]))
        )
    else:
        lines = (
            ([""] * padding[2])
            + lines
            + ([""] * (inner_dimensions[1] - len(lines) + padding[3]))
        )

    # Horizontally align and pad.
    for i, line in enumerate(lines):
        new_width = inner_dimensions[0] + len(line) - visible_width(line)
        if "right" in align:
            lines[i] = line.rjust(padding[0] + new_width, space) + (space * padding[1])
        elif "center" in align:
            lines[i] = (
                (space * padding[0])
                + line.center(new_width, space)
                + (space * padding[1])
            )
        else:
            lines[i] = (space * padding[0]) + line.ljust(new_width + padding[1], space)

    return lines


def max_dimensions(
    table_data, padding_left=0, padding_right=0, padding_top=0, padding_bottom=0
):
    """Get maximum widths of each column and maximum height of each row.

    :param iter table_data: List of list of strings (unmodified table data).
    :param int padding_left: Number of space chars on left side of cell.
    :param int padding_right: Number of space chars on right side of cell.
    :param int padding_top: Number of empty lines on top side of cell.
    :param int padding_bottom: Number of empty lines on bottom side of cell.

    :return: 4-item tuple of n-item lists. Inner column widths and row heights, outer column widths and row heights.
    :rtype: tuple
    """
    inner_widths = [0] * (max(len(r) for r in table_data) if table_data else 0)
    inner_heights = [0] * len(table_data)

    # Find max width and heights.
    for j, row in enumerate(table_data):
        for i, cell in enumerate(row):
            if not hasattr(cell, "count") or not hasattr(cell, "splitlines"):
                cell = str(cell)
            if not cell:
                continue
            inner_heights[j] = max(inner_heights[j], cell.count("\n") + 1)
            inner_widths[i] = max(
                inner_widths[i],
                *[visible_width(the_line) for the_line in cell.splitlines()]
            )

    # Calculate with padding.
    outer_widths = [padding_left + i + padding_right for i in inner_widths]
    outer_heights = [padding_top + i + padding_bottom for i in inner_heights]

    return inner_widths, inner_heights, outer_widths, outer_heights


def column_max_width(
    inner_widths: Sequence[int],
    column_number: int,
    outer_border: int,
    inner_border: int,
    padding: int,
) -> int:
    """Determine the maximum width of a column based on the current terminal width.

    :param iter inner_widths: List of widths (no padding) for each column.
    :param int column_number: The column number to query.
    :param int outer_border: Sum of left and right outer border visible widths.
    :param int inner_border: Visible width of the inner border character.
    :param int padding: Total padding per cell (left + right padding).

    :return: The maximum width the column can be without causing line wrapping.
    """
    column_count = len(inner_widths)
    terminal_width = terminal_size()[0]

    # Count how much space padding, outer, and inner borders take up.
    non_data_space = outer_border
    non_data_space += inner_border * (column_count - 1)
    non_data_space += column_count * padding

    # Exclude selected column's width.
    data_space = sum(inner_widths) - inner_widths[column_number]

    return terminal_width - data_space - non_data_space


def table_width(
    outer_widths: Sequence[int], outer_border: int, inner_border: int
) -> int:
    """Determine the width of the entire table including borders and padding.

    :param iter outer_widths: List of widths (with padding) for each column.
    :param int outer_border: Sum of left and right outer border visible widths.
    :param int inner_border: Visible width of the inner border character.

    :return: The width of the table.
    :rtype: int
    """
    column_count = len(outer_widths)

    # Count how much space outer and inner borders take up.
    non_data_space = outer_border
    if column_count:
        non_data_space += inner_border * (column_count - 1)

    # Space of all columns and their padding.
    data_space = sum(outer_widths)
    return data_space + non_data_space



"""Test method in BaseTable class."""

import pytest

from terminaltables3.base_table import BaseTable
from terminaltables3.build import flatten
from terminaltables3.width_and_alignment import max_dimensions


@pytest.mark.parametrize("inner_heading_row_border", [True, False])
@pytest.mark.parametrize("inner_footing_row_border", [True, False])
@pytest.mark.parametrize("inner_row_border", [True, False])
def test_inner_row_borders(
    inner_heading_row_border, inner_footing_row_border, inner_row_border
):
    """Test heading/footing/row borders.

    :param bool inner_heading_row_border: Passed to table.
    :param bool inner_footing_row_border: Passed to table.
    :param bool inner_row_border: Passed to table.
    """
    table_data = [
        ["Name", "Color", "Type"],
        ["Avocado", "green", "nut"],
        ["Tomato", "red", "fruit"],
        ["Lettuce", "green", "vegetable"],
    ]
    table = BaseTable(table_data)
    table.inner_heading_row_border = inner_heading_row_border
    table.inner_footing_row_border = inner_footing_row_border
    table.inner_row_border = inner_row_border
    inner_widths, inner_heights, outer_widths = max_dimensions(
        table_data, table.padding_left, table.padding_right
    )[:3]
    actual = flatten(table.gen_table(inner_widths, inner_heights, outer_widths))

    # Determine expected.
    if inner_row_border:
        expected = (
            "+---------+-------+-----------+\n"
            "| Name    | Color | Type      |\n"
            "+---------+-------+-----------+\n"
            "| Avocado | green | nut       |\n"
            "+---------+-------+-----------+\n"
            "| Tomato  | red   | fruit     |\n"
            "+---------+-------+-----------+\n"
            "| Lettuce | green | vegetable |\n"
            "+---------+-------+-----------+"
        )
    elif inner_heading_row_border and inner_footing_row_border:
        expected = (
            "+---------+-------+-----------+\n"
            "| Name    | Color | Type      |\n"
            "+---------+-------+-----------+\n"
            "| Avocado | green | nut       |\n"
            "| Tomato  | red   | fruit     |\n"
            "+---------+-------+-----------+\n"
            "| Lettuce | green | vegetable |\n"
            "+---------+-------+-----------+"
        )
    elif inner_heading_row_border:
        expected = (
            "+---------+-------+-----------+\n"
            "| Name    | Color | Type      |\n"
            "+---------+-------+-----------+\n"
            "| Avocado | green | nut       |\n"
            "| Tomato  | red   | fruit     |\n"
            "| Lettuce | green | vegetable |\n"
            "+---------+-------+-----------+"
        )
    elif inner_footing_row_border:
        expected = (
            "+---------+-------+-----------+\n"
            "| Name    | Color | Type      |\n"
            "| Avocado | green | nut       |\n"
            "| Tomato  | red   | fruit     |\n"
            "+---------+-------+-----------+\n"
            "| Lettuce | green | vegetable |\n"
            "+---------+-------+-----------+"
        )
    else:
        expected = (
            "+---------+-------+-----------+\n"
            "| Name    | Color | Type      |\n"
            "| Avocado | green | nut       |\n"
            "| Tomato  | red   | fruit     |\n"
            "| Lettuce | green | vegetable |\n"
            "+---------+-------+-----------+"
        )

    assert actual == expected


@pytest.mark.parametrize("outer_border", [True, False])
def test_outer_borders(outer_border):
    """Test left/right/top/bottom table borders.

    :param bool outer_border: Passed to table.
    """
    table_data = [
        ["Name", "Color", "Type"],
        ["Avocado", "green", "nut"],
        ["Tomato", "red", "fruit"],
        ["Lettuce", "green", "vegetable"],
    ]
    table = BaseTable(table_data, "Example Table")
    table.outer_border = outer_border
    inner_widths, inner_heights, outer_widths = max_dimensions(
        table_data, table.padding_left, table.padding_right
    )[:3]
    actual = flatten(table.gen_table(inner_widths, inner_heights, outer_widths))

    # Determine expected.
    if outer_border:
        expected = (
            "+Example Table----+-----------+\n"
            "| Name    | Color | Type      |\n"
            "+---------+-------+-----------+\n"
            "| Avocado | green | nut       |\n"
            "| Tomato  | red   | fruit     |\n"
            "| Lettuce | green | vegetable |\n"
            "+---------+-------+-----------+"
        )
    else:
        expected = (
            " Name    | Color | Type      \n"
            "---------+-------+-----------\n"
            " Avocado | green | nut       \n"
            " Tomato  | red   | fruit     \n"
            " Lettuce | green | vegetable "
        )

    assert actual == expected


@pytest.mark.parametrize("mode", ["row", "one", "blank", "empty", "none"])
@pytest.mark.parametrize("bare", [False, True])
def test_one_no_rows(mode, bare):
    """Test with one or no rows.

    :param str mode: Type of table contents to test.
    :param bool bare: Disable padding/borders.
    """
    if mode == "row":
        table_data = [
            ["Avocado", "green", "nut"],
        ]
    elif mode == "one":
        table_data = [
            ["Avocado"],
        ]
    elif mode == "blank":
        table_data = [
            [""],
        ]
    elif mode == "empty":
        table_data = [
            [],
        ]
    else:
        table_data = []
    table = BaseTable(table_data)
    if bare:
        table.inner_column_border = False
        table.inner_footing_row_border = False
        table.inner_heading_row_border = False
        table.inner_row_border = False
        table.outer_border = False
        table.padding_left = 0
        table.padding_right = 0
    inner_widths, inner_heights, outer_widths = max_dimensions(
        table_data, table.padding_left, table.padding_right
    )[:3]
    actual = flatten(table.gen_table(inner_widths, inner_heights, outer_widths))

    # Determine expected.
    if mode == "row":
        if bare:
            expected = "Avocadogreennut"
        else:
            expected = (
                "+---------+-------+-----+\n"
                "| Avocado | green | nut |\n"
                "+---------+-------+-----+"
            )
    elif mode == "one":
        if bare:
            expected = "Avocado"
        else:
            expected = "+---------+\n" "| Avocado |\n" "+---------+"
    elif mode == "blank":  # Remember there's still padding.
        if bare:
            expected = ""
        else:
            expected = "+--+\n" "|  |\n" "+--+"
    elif mode == "empty":
        if bare:
            expected = ""
        else:
            expected = "++\n" "||\n" "++"
    else:
        if bare:
            expected = ""
        else:
            expected = "++\n" "++"

    assert actual == expected



"""Test function in module."""

from terminaltables3.width_and_alignment import max_dimensions, table_width


def test_empty():
    """Test with zero-length cells."""
    assert table_width(max_dimensions([[""]])[2], 0, 0) == 0
    assert table_width(max_dimensions([["", "", ""]])[2], 0, 0) == 0
    assert table_width(max_dimensions([["", "", ""], ["", "", ""]])[2], 0, 0) == 0

    assert table_width(max_dimensions([[""]], 1, 1)[2], 2, 1) == 4
    assert table_width(max_dimensions([["", "", ""]], 1, 1)[2], 2, 1) == 10
    assert (
        table_width(max_dimensions([["", "", ""], ["", "", ""]], 1, 1)[2], 2, 1) == 10
    )


def test_single_line():
    """Test with single-line cells."""
    table_data = [
        ["Name", "Color", "Type"],
        ["Avocado", "green", "nut"],
        ["Tomato", "red", "fruit"],
        ["Lettuce", "green", "vegetable"],
    ]

    # '| Lettuce | green | vegetable |'
    outer, inner, outer_widths = 2, 1, max_dimensions(table_data, 1, 1)[2]
    assert table_width(outer_widths, outer, inner) == 31

    # ' Lettuce | green | vegetable '
    outer = 0
    assert table_width(outer_widths, outer, inner) == 29

    # '| Lettuce  green  vegetable |'
    outer, inner = 2, 0
    assert table_width(outer_widths, outer, inner) == 29

    # ' Lettuce  green  vegetable '
    outer = 0
    assert table_width(outer_widths, outer, inner) == 27

    # '|Lettuce |green |vegetable |'
    outer, inner, outer_widths = 2, 1, max_dimensions(table_data, 1)[2]
    assert table_width(outer_widths, outer, inner) == 28

    # '|Lettuce     |green     |vegetable     |'
    outer_widths = max_dimensions(table_data, 3, 2)[2]
    assert table_width(outer_widths, outer, inner) == 40

    table_data = [
        ["Name", "Color", "Type"],
        ["Avocado", "green", "nut"],
        ["Tomato", "red", "fruit"],
        ["Lettuce", "green", "vegetable"],
        ["Watermelon", "green", "fruit"],
    ]
    outer, inner, outer_widths = 2, 1, max_dimensions(table_data, 1, 1)[2]
    assert table_width(outer_widths, outer, inner) == 34


def test_multi_line():
    """Test with multi-line cells."""
    table_data = [
        ["Show", "Characters"],
        [
            "Rugrats",
            (
                "Tommy Pickles, Chuckie Finster, Phillip DeVille, Lillian DeVille, Angelica Pickles,\n"
                "Susie Carmichael, Dil Pickles, Kimi Finster, Spike"
            ),
        ],
        ["South Park", "Stan Marsh, Kyle Broflovski, Eric Cartman, Kenny McCormick"],
    ]
    outer, inner, outer_widths = 2, 1, max_dimensions(table_data, 1, 1)[2]
    assert table_width(outer_widths, outer, inner) == 100



"""Get info about the current terminal window/screen buffer."""

import ctypes
import struct
import sys
from typing import Tuple, Union

DEFAULT_HEIGHT = 24
DEFAULT_WIDTH = 79
INVALID_HANDLE_VALUE = -1
IS_WINDOWS = sys.platform == "win32"
STD_ERROR_HANDLE = -12
STD_OUTPUT_HANDLE = -11


def get_console_info(kernel32, handle: int) -> Tuple[int, int]:
    """Get information about this current console window (Windows only).

    https://github.com/Robpol86/colorclass/blob/ab42da59/colorclass/windows.py#L111

    :raise OSError: When handle is invalid or GetConsoleScreenBufferInfo API call fails.

    :param ctypes.windll.kernel32 kernel32: Loaded kernel32 instance.
    :param int handle: stderr or stdout handle.

    :return: Width (number of characters) and height (number of lines) of the terminal.
    :rtype: tuple
    """
    if handle == INVALID_HANDLE_VALUE:
        raise OSError("Invalid handle.")

    # Query Win32 API.
    lpcsbi = ctypes.create_string_buffer(22)  # Populated by GetConsoleScreenBufferInfo.
    if not kernel32.GetConsoleScreenBufferInfo(handle, lpcsbi):
        raise ctypes.WinError()  # Subclass of OSError.

    # Parse data.
    left, top, right, bottom = struct.unpack("hhhhHhhhhhh", lpcsbi.raw)[5:-2]
    width, height = right - left, bottom - top
    return width, height


def terminal_size(kernel32=None) -> Tuple[int, int]:
    """Get the width and height of the terminal.

    http://code.activestate.com/recipes/440694-determine-size-of-console-window-on-windows/
    http://stackoverflow.com/questions/17993814/why-the-irrelevant-code-made-a-difference

    :param kernel32: Optional mock kernel32 object. For testing.

    :return: Width (number of characters) and height (number of lines) of the terminal.
    :rtype: tuple
    """
    if IS_WINDOWS:
        kernel32 = kernel32 or ctypes.windll.kernel32
        try:
            return get_console_info(kernel32, kernel32.GetStdHandle(STD_ERROR_HANDLE))
        except OSError:
            try:
                return get_console_info(
                    kernel32, kernel32.GetStdHandle(STD_OUTPUT_HANDLE)
                )
            except OSError:
                return DEFAULT_WIDTH, DEFAULT_HEIGHT

    try:
        device = __import__("fcntl").ioctl(
            0, __import__("termios").TIOCGWINSZ, "\0\0\0\0\0\0\0\0"
        )
    except OSError:
        return DEFAULT_WIDTH, DEFAULT_HEIGHT
    height, width = struct.unpack("hhhh", device)[:2]
    return width, height


def set_terminal_title(title: Union[str, bytes], kernel32=None) -> bool:
    """Set the terminal title.

    :param title: The title to set (string, unicode, bytes accepted).
    :param kernel32: Optional mock kernel32 object. For testing.

    :return: If title changed successfully (Windows only, always True on Linux/OSX).
    :rtype: bool
    """
    try:
        title_bytes = title.encode("utf-8")
    except AttributeError:
        title_bytes = title

    if IS_WINDOWS:
        kernel32 = kernel32 or ctypes.windll.kernel32
        try:
            is_ascii = all(ord(c) < 128 for c in title)  # str/unicode.
        except TypeError:
            is_ascii = all(c < 128 for c in title)  # bytes.
        if is_ascii:
            return kernel32.SetConsoleTitleA(title_bytes) != 0
        return kernel32.SetConsoleTitleW(title) != 0

    # Linux/OSX.
    sys.stdout.write(b"\033]0;" + title_bytes + b"\007")
    return True



"""Test function in module."""

from terminaltables3.build import build_row


def test_one_line():
    """Test with one line cells."""
    row = [
        ["Left Cell"],
        ["Center Cell"],
        ["Right Cell"],
    ]
    actual = [tuple(i) for i in build_row(row, ">", "|", "<")]
    expected = [
        (">", "Left Cell", "|", "Center Cell", "|", "Right Cell", "<"),
    ]
    assert actual == expected


def test_two_line():
    """Test with two line cells."""
    row = [
        [
            "Left ",
            "Cell1",
        ],
        [
            "Center",
            "Cell2 ",
        ],
        [
            "Right",
            "Cell3",
        ],
    ]
    actual = [tuple(i) for i in build_row(row, ">", "|", "<")]
    expected = [
        (">", "Left ", "|", "Center", "|", "Right", "<"),
        (">", "Cell1", "|", "Cell2 ", "|", "Cell3", "<"),
    ]
    assert actual == expected


def test_three_line():
    """Test with three line cells."""
    row = [
        [
            "Left ",
            "Cell1",
            "     ",
        ],
        [
            "Center",
            "Cell2 ",
            "      ",
        ],
        [
            "Right",
            "Cell3",
            "     ",
        ],
    ]
    actual = [tuple(i) for i in build_row(row, ">", "|", "<")]
    expected = [
        (">", "Left ", "|", "Center", "|", "Right", "<"),
        (">", "Cell1", "|", "Cell2 ", "|", "Cell3", "<"),
        (">", "     ", "|", "      ", "|", "     ", "<"),
    ]
    assert actual == expected


def test_single():
    """Test with single cell."""
    actual = [tuple(i) for i in build_row([["Cell"]], ">", "|", "<")]
    expected = [
        (">", "Cell", "<"),
    ]
    assert actual == expected


def test_empty():
    """Test with empty cell."""
    actual = [tuple(i) for i in build_row([[""]], ">", "|", "<")]
    expected = [
        (">", "", "<"),
    ]
    assert actual == expected


def test_no_cells():
    """Test with no cells."""
    actual = [tuple(i) for i in build_row([[]], ">", "|", "<")]
    expected = [
        (">", "<"),
    ]
    assert actual == expected

    actual = [tuple(i) for i in build_row([], ">", "|", "<")]
    expected = [
        (">", "<"),
    ]
    assert actual == expected



#!/usr/bin/env python
"""Simple example usage of terminaltables without any other dependencies.

Just prints sample text and exits.
"""


from terminaltables3 import AsciiTable, DoubleTable, SingleTable

TABLE_DATA = (
    ("Platform", "Years", "Notes"),
    ("Mk5", "2007-2009", "The Golf Mk5 Variant was\nintroduced in 2007."),
    ("MKVI", "2009-2013", "Might actually be Mk5."),
)


def main():
    """Main function."""
    title = "Jetta SportWagen"

    # AsciiTable.
    table_instance = AsciiTable(TABLE_DATA, title)
    table_instance.justify_columns[2] = "right"
    print(table_instance.table)
    print()

    # SingleTable.
    table_instance = SingleTable(TABLE_DATA, title)
    table_instance.justify_columns[2] = "right"
    print(table_instance.table)
    print()

    # DoubleTable.
    table_instance = DoubleTable(TABLE_DATA, title)
    table_instance.justify_columns[2] = "right"
    print(table_instance.table)
    print()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""Example usage of terminaltables with colorclass.

Just prints sample text and exits.
"""


from colorclass import Color, Windows

from terminaltables3 import SingleTable


def table_server_timings():
    """Return table string to be printed."""
    table_data = [
        [Color("{autogreen}<10ms{/autogreen}"), "192.168.0.100, 192.168.0.101"],
        [
            Color("{autoyellow}10ms <= 100ms{/autoyellow}"),
            "192.168.0.102, 192.168.0.103",
        ],
        [Color("{autored}>100ms{/autored}"), "192.168.0.105"],
    ]
    table_instance = SingleTable(table_data)
    table_instance.inner_heading_row_border = False
    return table_instance.table


def table_server_status():
    """Return table string to be printed."""
    table_data = [
        [
            Color("Low Space"),
            Color("{autocyan}Nominal Space{/autocyan}"),
            Color("Excessive Space"),
        ],
        [
            Color("Low Load"),
            Color("Nominal Load"),
            Color("{autored}High Load{/autored}"),
        ],
        [
            Color("{autocyan}Low Free RAM{/autocyan}"),
            Color("Nominal Free RAM"),
            Color("High Free RAM"),
        ],
    ]
    table_instance = SingleTable(table_data, "192.168.0.105")
    table_instance.inner_heading_row_border = False
    table_instance.inner_row_border = True
    table_instance.justify_columns = {0: "center", 1: "center", 2: "center"}
    return table_instance.table


def table_abcd():
    """Return table string to be printed. Two tables on one line."""
    table_instance = SingleTable([["A", "B"], ["C", "D"]])

    # Get first table lines.
    table_instance.outer_border = False
    table_inner_borders = table_instance.table.splitlines()

    # Get second table lines.
    table_instance.outer_border = True
    table_instance.inner_heading_row_border = False
    table_instance.inner_column_border = False
    table_outer_borders = table_instance.table.splitlines()

    # Combine.
    smallest, largest = sorted([table_inner_borders, table_outer_borders], key=len)
    smallest += [""] * (len(largest) - len(smallest))  # Make both same size.
    combined = []
    for i, row in enumerate(largest):
        combined.append(row.ljust(10) + "          " + smallest[i])
    return "\n".join(combined)


def main():
    """Main function."""
    Windows.enable(
        auto_colors=True, reset_atexit=True
    )  # Does nothing if not on Windows.

    # Server timings.
    print(table_server_timings())
    print()

    # Server status.
    print(table_server_status())
    print()

    # Two A B C D tables.
    print(table_abcd())
    print()

    # Instructions.
    table_instance = SingleTable([["Obey Obey Obey Obey"]], "Instructions")
    print(table_instance.table)
    print()


if __name__ == "__main__":
    main()



"""Common objects used by tests in directory."""

from terminaltables3 import terminal_io


class MockKernel32:
    """Mock kernel32."""

    def __init__(
        self,
        stderr=terminal_io.INVALID_HANDLE_VALUE,
        stdout=terminal_io.INVALID_HANDLE_VALUE,
    ):
        """Constructor."""
        self.stderr = stderr
        self.stdout = stdout
        self.csbi_err = b"x\x00)#\x00\x00\x87\x05\x07\x00\x00\x00j\x05w\x00\x87\x05x\x00J\x00"  # 119 x 29
        self.csbi_out = b"L\x00,\x01\x00\x00*\x01\x07\x00\x00\x00\x0e\x01K\x00*\x01L\x00L\x00"  # 75 x 28
        self.setConsoleTitleA_called = False
        self.setConsoleTitleW_called = False

    def GetConsoleScreenBufferInfo(self, handle, lpcsbi):  # noqa
        """Mock GetConsoleScreenBufferInfo.

        :param handle: Unused handle.
        :param lpcsbi: ctypes.create_string_buffer() return value.
        """
        if handle == self.stderr:
            lpcsbi.raw = self.csbi_err
        else:
            lpcsbi.raw = self.csbi_out
        return 1

    def GetStdHandle(self, handle):  # noqa
        """Mock GetStdHandle.

        :param int handle: STD_ERROR_HANDLE or STD_OUTPUT_HANDLE.
        """
        return self.stderr if handle == terminal_io.STD_ERROR_HANDLE else self.stdout

    def SetConsoleTitleA(self, _):  # noqa
        """Mock SetConsoleTitleA."""
        self.setConsoleTitleA_called = True
        return 1

    def SetConsoleTitleW(self, _):  # noqa
        """Mock SetConsoleTitleW."""
        self.setConsoleTitleW_called = True
        return 1



"""Test function in module."""

import pytest

from terminaltables3.width_and_alignment import column_max_width, max_dimensions


@pytest.fixture(autouse=True)
def patch(monkeypatch):
    """Monkeypatch before every test function in this module.

    :param monkeypatch: pytest fixture.
    """
    monkeypatch.setattr(
        "terminaltables3.width_and_alignment.terminal_size", lambda: (79, 24)
    )


def test_empty():
    """Test with zero-length cells."""
    assert column_max_width(max_dimensions([[""]])[0], 0, 0, 0, 0) == 79
    assert column_max_width(max_dimensions([["", "", ""]])[0], 0, 0, 0, 0) == 79
    assert (
        column_max_width(max_dimensions([["", "", ""], ["", "", ""]])[0], 0, 0, 0, 0)
        == 79
    )

    assert column_max_width(max_dimensions([[""]])[0], 0, 2, 1, 2) == 75
    assert column_max_width(max_dimensions([["", "", ""]])[0], 0, 2, 1, 2) == 69
    assert (
        column_max_width(max_dimensions([["", "", ""], ["", "", ""]])[0], 0, 2, 1, 2)
        == 69
    )


def test_single_line():
    """Test with single-line cells."""
    table_data = [
        ["Name", "Color", "Type"],
        ["Avocado", "green", "nut"],
        ["Tomato", "red", "fruit"],
        ["Lettuce", "green", "vegetable"],
    ]
    inner_widths = max_dimensions(table_data)[0]

    # '| Lettuce | green | vegetable |'
    outer, inner, padding = 2, 1, 2
    assert column_max_width(inner_widths, 0, outer, inner, padding) == 55
    assert column_max_width(inner_widths, 1, outer, inner, padding) == 53
    assert column_max_width(inner_widths, 2, outer, inner, padding) == 57

    # ' Lettuce | green | vegetable '
    outer = 0
    assert column_max_width(inner_widths, 0, outer, inner, padding) == 57
    assert column_max_width(inner_widths, 1, outer, inner, padding) == 55
    assert column_max_width(inner_widths, 2, outer, inner, padding) == 59

    # '| Lettuce  green  vegetable |'
    outer, inner = 2, 0
    assert column_max_width(inner_widths, 0, outer, inner, padding) == 57
    assert column_max_width(inner_widths, 1, outer, inner, padding) == 55
    assert column_max_width(inner_widths, 2, outer, inner, padding) == 59

    # ' Lettuce  green  vegetable '
    outer = 0
    assert column_max_width(inner_widths, 0, outer, inner, padding) == 59
    assert column_max_width(inner_widths, 1, outer, inner, padding) == 57
    assert column_max_width(inner_widths, 2, outer, inner, padding) == 61

    # '|Lettuce |green |vegetable |'
    outer, inner, padding = 2, 1, 1
    assert column_max_width(inner_widths, 0, outer, inner, padding) == 58
    assert column_max_width(inner_widths, 1, outer, inner, padding) == 56
    assert column_max_width(inner_widths, 2, outer, inner, padding) == 60

    # '|Lettuce     |green     |vegetable     |'
    padding = 5
    assert column_max_width(inner_widths, 0, outer, inner, padding) == 46
    assert column_max_width(inner_widths, 1, outer, inner, padding) == 44
    assert column_max_width(inner_widths, 2, outer, inner, padding) == 48

    table_data = [
        ["Name", "Color", "Type"],
        ["Avocado", "green", "nut"],
        ["Tomato", "red", "fruit"],
        ["Lettuce", "green", "vegetable"],
        ["Watermelon", "green", "fruit"],
    ]
    inner_widths = max_dimensions(table_data)[0]
    outer, inner, padding = 2, 1, 2
    assert column_max_width(inner_widths, 0, outer, inner, padding) == 55
    assert column_max_width(inner_widths, 1, outer, inner, padding) == 50
    assert column_max_width(inner_widths, 2, outer, inner, padding) == 54


def test_multi_line(monkeypatch):
    """Test with multi-line cells.

    :param monkeypatch: pytest fixture.
    """
    table_data = [
        ["Show", "Characters"],
        [
            "Rugrats",
            (
                "Tommy Pickles, Chuckie Finster, Phillip DeVille, Lillian DeVille, Angelica Pickles,\n"
                "Susie Carmichael, Dil Pickles, Kimi Finster, Spike"
            ),
        ],
        ["South Park", "Stan Marsh, Kyle Broflovski, Eric Cartman, Kenny McCormick"],
    ]
    inner_widths = max_dimensions(table_data)[0]
    outer, inner, padding = 2, 1, 2

    assert column_max_width(inner_widths, 0, outer, inner, padding) == -11
    assert column_max_width(inner_widths, 1, outer, inner, padding) == 62

    monkeypatch.setattr(
        "terminaltables3.width_and_alignment.terminal_size", lambda: (100, 24)
    )
    assert column_max_width(inner_widths, 0, outer, inner, padding) == 10
    assert column_max_width(inner_widths, 1, outer, inner, padding) == 83



"""Test function in module."""

import pytest

from terminaltables3.build import combine


@pytest.mark.parametrize("generator", [False, True])
def test_borders(generator):
    """Test with borders.

    :param bool generator: Test with generator instead of list.
    """
    line = ["One", "Two", "Three"]
    actual = list(combine(iter(line) if generator else line, ">", "|", "<"))
    assert actual == [">", "One", "|", "Two", "|", "Three", "<"]


@pytest.mark.parametrize("generator", [False, True])
def test_no_border(generator):
    """Test without borders.

    :param bool generator: Test with generator instead of list.
    """
    line = ["One", "Two", "Three"]
    actual = list(combine(iter(line) if generator else line, "", "", ""))
    assert actual == ["One", "Two", "Three"]


@pytest.mark.parametrize("generator", [False, True])
def test_no_items(generator):
    """Test with empty list.

    :param bool generator: Test with generator instead of list.
    """
    actual = list(combine(iter([]) if generator else [], ">", "|", "<"))
    assert actual == [">", "<"]



"""Test function in module."""

import sys
from textwrap import dedent

import py
import pytest

from terminaltables3.terminal_io import IS_WINDOWS, set_terminal_title

from tests import PROJECT_ROOT
from tests.screenshot import RunNewConsole, screenshot_until_match
from tests.test_terminal_io import MockKernel32

HERE = py.path.local(__file__).dirpath()


@pytest.mark.skip("Fails on windows, I didn't touch it")
@pytest.mark.parametrize("is_windows", [False, True])
@pytest.mark.parametrize("mode", ["ascii", "unicode", "bytes"])
def test(monkeypatch, is_windows, mode):
    """Test function.

    :param monkeypatch: pytest fixture.
    :param bool is_windows: Monkeypatch terminal_io.IS_WINDOWS
    :param str mode: Scenario to test for.
    """
    monkeypatch.setattr("terminaltables3.terminal_io.IS_WINDOWS", is_windows)
    kernel32 = MockKernel32()

    # Title.
    if mode == "ascii":
        title = "Testing terminaltables3."
    elif mode == "unicode":
        title = "Testing terminaltables3 with unicode: 世界你好蓝色"
    else:
        title = b"Testing terminaltables3 with bytes."

    # Run.
    assert set_terminal_title(title, kernel32)
    if not is_windows:
        return

    # Verify.
    if mode == "ascii":
        assert kernel32.setConsoleTitleA_called
        assert not kernel32.setConsoleTitleW_called
    elif mode == "unicode":
        assert not kernel32.setConsoleTitleA_called
        assert kernel32.setConsoleTitleW_called
    else:
        assert kernel32.setConsoleTitleA_called
        assert not kernel32.setConsoleTitleW_called


@pytest.mark.skipif(str(not IS_WINDOWS))
@pytest.mark.parametrize("mode", ["ascii", "unicode", "bytes"])
@pytest.mark.skip  # https://github.com/Robpol86/terminaltables3/issues/44
def test_windows_screenshot(tmpdir, mode):
    """Test function on Windows in a new console window. Take a screenshot to verify it works.

    :param tmpdir: pytest fixture.
    :param str mode: Scenario to test for.
    """
    script = tmpdir.join("script.py")
    command = [sys.executable, str(script)]
    change_title = tmpdir.join("change_title")
    screenshot = PROJECT_ROOT.join(f"test_terminal_io_{mode}.png")
    if screenshot.check():
        screenshot.remove()

    # Determine title.
    if mode == "ascii":
        title = "'test ASCII test'"
    elif mode == "unicode":
        title = "u'test 世界你好蓝色 test'"
    else:
        title = "b'test ASCII test'"

    # Generate script.
    script_template = dedent(
        """\
    # coding: utf-8
    from __future__ import print_function
    import os, time
    from terminaltables3.terminal_io import set_terminal_title
    stop_after = time.time() + 20

    print('Waiting for FindWindowA() in RunNewConsole.__enter__()...')
    while not os.path.exists(r'{change_title}') and time.time() < stop_after:
        time.sleep(0.5)
    assert set_terminal_title({title}) is True

    print('Waiting for screenshot_until_match()...')
    while not os.path.exists(r'{screenshot}') and time.time() < stop_after:
        time.sleep(0.5)
    """
    )
    script_contents = script_template.format(
        change_title=str(change_title), title=title, screenshot=str(screenshot)
    )
    script.write(script_contents.encode("utf-8"), mode="wb")

    # Setup expected.
    if mode == "unicode":
        sub_images = [str(p) for p in HERE.listdir("sub_title_cjk_*.bmp")]
    else:
        sub_images = [str(p) for p in HERE.listdir("sub_title_ascii_*.bmp")]
    assert sub_images

    # Run.
    with RunNewConsole(command) as gen:
        change_title.ensure(file=True)  # Touch file.
        screenshot_until_match(str(screenshot), 15, sub_images, 1, gen)
